#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: Enthought, Inc.
#
#------------------------------------------------------------------------------

""" The wx specific implementation of the tool bar manager.
"""

# Major package imports.
import wx

# Enthought library imports.
from traits.api import Bool, Enum, Instance, Str, Tuple

# Local imports.
from pyface.wx.aui import aui
from pyface.image_cache import ImageCache
from pyface.action.action_manager import ActionManager


class ToolBarManager(ActionManager):
    """ A tool bar manager realizes itself in errr, a tool bar control. """

    #### 'ToolBarManager' interface ###########################################

    # Is the tool bar enabled?
    enabled = Bool(True)

    # Is the tool bar visible?
    visible = Bool(True)

    # The size of tool images (width, height).
    image_size = Tuple((16, 16))

    # The toolbar name (used to distinguish multiple toolbars).
    name = Str('ToolBar')

    # The orientation of the toolbar.
    orientation = Enum('horizontal', 'vertical')

    # Should we display the name of each tool bar tool under its image?
    show_tool_names = Bool(True)

    # Should we display the horizontal divider?
    show_divider = Bool(False)

    #### Private interface ####################################################

    # Cache of tool images (scaled to the appropriate size).
    _image_cache = Instance(ImageCache)

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, *args, **traits):
        """ Creates a new tool bar manager. """

        # Base class contructor.
        super(ToolBarManager, self).__init__(*args, **traits)

        # An image cache to make sure that we only load each image used in the
        # tool bar exactly once.
        self._image_cache = ImageCache(self.image_size[0], self.image_size[1])

        return

    ###########################################################################
    # 'ToolBarManager' interface.
    ###########################################################################

    #### Trait change handlers ################################################
    #### Methods ##############################################################

    def create_tool_bar(self, parent, controller=None, aui=False):
        """ Creates a tool bar. """

        # If a controller is required it can either be set as a trait on the
        # tool bar manager (the trait is part of the 'ActionManager' API), or
        # passed in here (if one is passed in here it takes precedence over the
        # trait).
        if controller is None:
            controller = self.controller

        # Determine the wx style for the tool bar based on any optional
        # settings.
        style = wx.NO_BORDER | wx.TB_FLAT | wx.CLIP_CHILDREN

        if self.show_tool_names:
            style |= wx.TB_TEXT

        if self.orientation == 'horizontal':
            style |= wx.TB_HORIZONTAL

        else:
            style |= wx.TB_VERTICAL

        if not self.show_divider:
            style |= wx.TB_NODIVIDER

        # Create the control.
        if aui:
            tool_bar = _AuiToolBar(self, parent, -1, style=style)
        else:
            tool_bar = _ToolBar(self, parent, -1, style=style)

        # fixme: Setting the tool bitmap size seems to be the only way to
        # change the height of the toolbar in wx.
        tool_bar.SetToolBitmapSize(self.image_size)

        # Add all of items in the manager's groups to the tool bar.
        self._wx_add_tools(parent, tool_bar, controller)

        # Make the tools appear in the tool bar (without this you will see
        # nothing!).
        tool_bar.Realize()

        # fixme: Without the following hack,  only the first item in a radio
        # group can be selected when the tool bar is first realised 8^()
        self._wx_set_initial_tool_state(tool_bar)

        return tool_bar

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _wx_add_tools(self, parent, tool_bar, controller):
        """ Adds tools for all items in the list of groups. """

        previous_non_empty_group = None
        for group in self.groups:
            if len(group.items) > 0:
                # Is a separator required?
                if previous_non_empty_group is not None and group.separator:
                    tool_bar.AddSeparator()

                previous_non_empty_group = group

                # Create a tool bar tool for each item in the group.
                for item in group.items:
                    item.add_to_toolbar(
                        parent,
                        tool_bar,
                        self._image_cache,
                        controller,
                        self.show_tool_names
                    )

        return

    def _wx_set_initial_tool_state(self, tool_bar):
        """ Workaround for the wxPython tool bar bug.

        Without this,  only the first item in a radio group can be selected
         when the tool bar is first realised 8^()

        """

        for group in self.groups:
            checked = False
            for item in group.items:
                # If the group is a radio group,  set the initial checked state
                # of every tool in it.
                if item.action.style == 'radio':
                    if item.control_id is not None:
                        # Only set checked state if control has been created.
                        # Using extra_actions of tasks, it appears that this
                        # may be called multiple times.
                        tool_bar.ToggleTool(item.control_id, item.action.checked)
                        checked = checked or item.action.checked

                # Every item in a radio group MUST be 'radio' style, so we
                # can just skip to the next group.
                else:
                    break

            # We get here if the group is a radio group.
            else:
                # If none of the actions in the group is specified as 'checked'
                # we will check the first one.
                if not checked and len(group.items) > 0:
                    group.items[0].action.checked = True

        return


class _ToolBar(wx.ToolBar):
    """ The toolkit-specific tool bar implementation. """

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, tool_bar_manager, parent, id, style):
        """ Constructor. """

        wx.ToolBar.__init__(self, parent, -1, style=style)

        # Listen for changes to the tool bar manager's enablement and
        # visibility.
        self.tool_bar_manager = tool_bar_manager

        self.tool_bar_manager.on_trait_change(
            self._on_tool_bar_manager_enabled_changed, 'enabled'
        )

        self.tool_bar_manager.on_trait_change(
            self._on_tool_bar_manager_visible_changed, 'visible'
        )

        return

    ###########################################################################
    # Trait change handlers.
    ###########################################################################

    def _on_tool_bar_manager_enabled_changed(self, obj, trait_name, old, new):
        """ Dynamic trait change handler. """

        obj.window._wx_enable_tool_bar(self, new)

        return

    def _on_tool_bar_manager_visible_changed(self, obj, trait_name, old, new):
        """ Dynamic trait change handler. """

        obj.window._wx_show_tool_bar(self, new)

        return


class _AuiToolBar(aui.AuiToolBar):
    """ The toolkit-specific tool bar implementation for AUI windows. """

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, tool_bar_manager, parent, id, style):
        """ Constructor. """

        aui.AuiToolBar.__init__(self, parent, -1, style=style)

        # Listen for changes to the tool bar manager's enablement and
        # visibility.
        self.tool_bar_manager = tool_bar_manager

        self.tool_bar_manager.on_trait_change(
            self._on_tool_bar_manager_enabled_changed, 'enabled'
        )

        self.tool_bar_manager.on_trait_change(
            self._on_tool_bar_manager_visible_changed, 'visible'
        )

        # we need to defer hiding tools until first time Realize is called so
        # we can get the correct order of the toolbar for reinsertion at the
        # correct position
        self.initially_hidden_tool_ids = []

        # map of tool ids to a tuple: position in full toolbar and the
        # ToolBarTool itself.  Can't keep a weak reference here because once
        # removed from the toolbar the item would be garbage collected.
        self.tool_map = {}

        return

    def Realize(self):
        if len(self.tool_map) == 0:
            for pos in range(self.GetToolsCount()):
                tool = self.GetToolByPos(pos)
                self.tool_map[tool.GetId()] = (pos, tool)
        aui.AuiToolBar.Realize(self)
        if len(self.initially_hidden_tool_ids) > 0:
            for tool_id in self.initially_hidden_tool_ids:
                self.RemoveTool(tool_id)
            self.initially_hidden_tool_ids = []
        self.ShowTool = self.ShowToolPostRealize

    def ShowTool(self, tool_id, state):
        """Used before realization to flag which need to be initially hidden
        """
        if not state:
            self.initially_hidden_tool_ids.append(tool_id)

    def ShowToolPostRealize(self, tool_id, state):
        """Normal ShowTool method, activated after first call to Realize
        """
        tool = self.FindById(tool_id)
        if state and tool is None:
            self.InsertToolInOrder(tool_id)
            self.EnableTool(tool_id, True)
            self.Realize()
            # Update the toolbar in the AUI manager to force toolbar resize
            wx.CallAfter(self.tool_bar_manager.controller.task.window._aui_manager.Update)
        elif not state and tool is not None:
            self.RemoveTool(tool_id)
            # Update the toolbar in the AUI manager to force toolbar resize
            wx.CallAfter(self.tool_bar_manager.controller.task.window._aui_manager.Update)

    def InsertToolInOrder(self, tool_id):
        orig_pos, tool = self.tool_map[tool_id]
        for pos in range(self.GetToolsCount()):
            existing_tool = self.GetToolByPos(pos)
            existing_id = existing_tool.GetId()
            existing_orig_pos, _ = self.tool_map[tool_id]
            if existing_orig_pos > orig_pos:
                break
        self.InsertToolItem(pos+1, tool)


    ##### Additional convenience functions for the normal AGW AUI toolbar

    def AddLabelTool(self, id, label, bitmap, bmpDisabled, kind, shortHelp,
                     longHelp, clientData):
        "The full AddTool() function."
        return self.AddTool(id, label, bitmap, bmpDisabled, kind, shortHelp,
                            longHelp, clientData, None)

    def InsertToolItem(self, pos, tool):
        self._items[pos:pos] = [tool]
        return tool

    def DeleteTool(self, tool_id):
        """
        Removes the specified tool from the toolbar and deletes it.

        :param integer `tool_id`: the :class:`AuiToolBarItem` identifier.
        :returns: ``True`` if the tool was deleted, ``False`` otherwise.
        :note: Note that it is unnecessary to call :meth:`Realize` for the
            change to take place, it will happen immediately.
        """

        tool = self.RemoveTool(tool_id)
        if tool is not None:
            tool.Destroy()
            return True

        return False


    def RemoveTool(self, tool_id):
        """
        Removes the specified tool from the toolbar but doesn't delete it.

        :param integer `tool_id`: the :class:`AuiToolBarItem` identifier.
        :returns: ``True`` if the tool was deleted, ``False`` otherwise.
        :note: Note that it is unnecessary to call :meth:`Realize` for the
            change to take place, it will happen immediately.
        """

        idx = self.GetToolIndex(tool_id)

        if idx >= 0 and idx < len(self._items):
            self._items.pop(idx)
            self.Realize()
            return True

        return False

    FindById = aui.AuiToolBar.FindTool

    GetToolState = aui.AuiToolBar.GetToolToggled

    GetToolsCount = aui.AuiToolBar.GetToolCount

    def GetToolByPos(self, pos):
        return self._items[pos]

    def OnSize(self, event):
        # Quickly short-circuit if the toolbar isn't realized
        if not hasattr(self, '_absolute_min_size'):
            return

        aui.AuiToolBar.OnSize(self, event)


    ###########################################################################
    # Trait change handlers.
    ###########################################################################

    def _on_tool_bar_manager_enabled_changed(self, obj, trait_name, old, new):
        """ Dynamic trait change handler. """

        obj.controller.task.window._wx_enable_tool_bar(self, new)

        return

    def _on_tool_bar_manager_visible_changed(self, obj, trait_name, old, new):
        """ Dynamic trait change handler. """

        obj.controller.task.window._wx_show_tool_bar(self, new)

        return

#### EOF ######################################################################



# peppy Copyright (c) 2006-2010 Rob McMullen
# Licenced under the GPLv2; see http://peppy.flipturn.org for more info

import numpy as np

from stcinterface import STCInterface, STCBinaryMixin


class BinarySTC(STCInterface, STCBinaryMixin):
    """
    Methods that a data source object must implement in order to be
    compatible with the real STC used as the data source for
    text-based files.

    See U{the Yellowbrain guide to the
    STC<http://www.yellowbrain.com/stc/index.html>} for more info on
    the rest of the STC methods.
    """

    def __init__(self):
        self.data = None

    def GetReadOnly(self):
        """Is the instance read-only (non-editable) or editable?"""
        return False

    def CanSave(self):
        """Can this STC instance save its contents?"""
        return True

    def Clear(self):
        pass

    def CanCopy(self):
        return False

    def Copy(self):
        pass

    def CanCut(self):
        return False

    def Cut(self):
        pass

    def CanPaste(self):
        return False

    def Paste(self):
        pass

    def EmptyUndoBuffer(self):
        pass

    def CanUndo(self):
        return False

    def Undo(self):
        pass

    def CanRedo(self):
        return False

    def Redo(self):
        pass

    def SetSavePoint(self):
        pass

    def GetText(self):
        return ''

    def GetLength(self):
        if self.data is not None:
            return self.data.size
        return 0

    GetTextLength = GetLength

    def GetModify(self):
        return False

    def CreateDocument(self):
        return "notarealdoc"

    def SetDocPointer(self,ptr):
        pass

    def ReleaseDocument(self,ptr):
        pass

    def AddRefDocument(self,ptr):
        pass

    def GuessBinary(self,amount,percentage):
        return False

    def getShortDisplayName(self, url):
        """Return a short name for display in tabs or other context without
        needing a pathname.
        """
        return url.path.get_name()

    def open(self, buffer, message=None):
        """Read from the specified url to populate the STC.
        
        Abstract method that subclasses use to read data into the STC.  May be
        called from a thread, so don't process any GUI actions here.

        buffer: buffer object used to read the file
        
        message: optional message used to update a progress bar
        """
        pass

    def revertEncoding(self, buffer, url=None, message=None, encoding=None, allow_undo=False):
        """Revert the file to the last saved state.
        
        @param buffer: buffer object used to read the file
        
        @param url: optional alternate URL, which if present means that the
        file should be reverted from the this URL instead of the original one.
        
        @param message: optional message used to update a progress bar
        
        @param encoding: optional encoding to attempt to convert from
        
        @param allow_undo: optional argument that if True and the STC is capable
        of it, allows the revert operation to be undoable.  If False or the
        STC isn't capable of undoing a revert, this parameter is ignored.
        """
        pass

    def readThreaded(self, fh, buffer, message=None):
        """Read from filehandle, converting as necessary.
        
        This may be called from a background thread, so no direct interaction
        with the GUI is allowed.  Any communication for progress bars needs to
        be done through the message that's passed in, using pubsub like::
        
            Publisher().sendMessage(message, percent)
        
        where percent is an integer from 0 to 100 indicating the amount of the
        file that has been successfully processed.  A value of -1 can be used
        to indicate that an unknown amount of data remains.

        @param fh: file-like object used to load the file
        @param buffer: L{Buffer} object containing information about the file
        @param message: optional pubsub message to be sent with progress
        """
        pass

    def openSuccess(self, buffer):
        """Called after a file has been successfully opened.
        
        This is called by the GUI thread, so can update any GUI elements here.
        """
        pass

    def getAutosaveTemporaryFilename(self, buffer):
        """Hook to allow STC to specify autosave filename"""
        pass

    def getBackupTemporaryFilename(self, buffer):
        """Hook to allow STC to override backup filename"""
        pass

    def prepareEncoding(self):
        """Convert the raw bytes in the file to the correct encoding before
        writing.
        
        """
        pass

    def openFileForWriting(self, url):
        """Return a file handle that has been opened for writing"""
        return None

    def writeTo(self, fh, url):
        """Write to filehandle, converting as necessary

        @param fh: file-like object to which the data should be saved
        @param url: the url that was used to open the file-like object
        """
        pass

    def closeFileAfterWriting(self, fh):
        """Close the opened file handle and perform any other cleanup"""
        pass

    def getProperties(self):
        """Return a list of properties to be displayed as text to the user
        
        @return: list of (name, value) pairs
        """
        return []

    def showStyle(self, linenum=None):
        """Debugging routine to show the styling information on a line.

        Print styling information to stdout to aid in debugging.
        """
        pass

    def GetFoldLevel(self, line):
        """Return fold level of specified line.

        Return fold level of line, which seems to be the number of spaces used
        to indent the line, plus an offset and shifted by 2^10
        """
        return 1024

    def GetFoldColumn(self, line):
        """Return column number of folding.

        Return column number of the current fold level.
        """
        return 0

    def GetPrevLineIndentation(self, line):
        """Get the indentation of the line before the specified line.

        Return a tuple containing the number of columns of indentation
        of the first non-blank line before the specified line, and the
        line number of the line that it found.
        """
        return 0, -1

    def GotoPos(self, pos):
        """Move the cursor to the specified position and scroll the
        position into the view if necessary.
        """
        pass

    def addUpdateUIEvent(self, callback):
        """Add the equivalent to STC_UPDATEUI event for UI changes.

        The STC supplies the EVT_STC_UPDATEUI event that fires for
        every change that could be used to update the user interface:
        a text change, a style change, or a selection change.  If the
        editing (viewing) window does not use the STC to display
        information, you should supply the equivalent event for the
        edit window.
        
        @param callback: event handler to execute on event
        """
        pass

    def addDocumentChangeEvent(self, callback):
        """Add the equivalent to EVT_STC_CHANGE event for document changes.

        The STC supplies the EVT_STC_CHANGE event that fires for every change
        to the contents of the document.  If you want peppy to be able to show
        indications that the document has been modified (which also allows
        peppy to display a dialog prompting for unsaved changes), register an
        event handler here that calls the specified callback when the document
        is changed.
        
        @param callback: event handler to execute on event
        """
        pass

    def removeDocumentChangeEvent(self):
        """Remove the document change event.
        
        The complement of L{addDocumentChangeEvent}, this is needed for example
        during revert processing so that all the intermediate changes don't
        generate change events that could propagate to user-level callbacks
        like the code explorer.
        """
        pass

    def GetBytes(self, start, end=-1):
        """Return the raw bytes between the given locations.
        
        @param start: starting offset
        @param end: ending offset, passing -1 means end of file
        """
        return self.data[start:end].tostring()

    def SetBytes(self, start, end, bytes):
        """Set raw bytes between the given locations.
        
        Note that bytes can be inserted or deleted if the length of the new
        byte string is different than the specified range.
        
        @param start: starting offset
        @param end: ending offset
        @param bytes: new bytes to replace existing data
        """
        bytes = np.fromstring(bytes, dtype=np.uint8)
        self.data[start:end] = bytes

    def SetBinary(self, data):
        self.data = np.fromstring(data, dtype=np.uint8)
        self.EmptyUndoBuffer()



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/29/2004
#
#------------------------------------------------------------------------------

""" Defines a wxPython ImageControl widget that is used by various trait
    editors to display trait values iconically.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

import wx

#-------------------------------------------------------------------------
#  'ImageControl' class:
#-------------------------------------------------------------------------


class ImageControl(wx.Window):
    """ A wxPython control that displays an image, which can be selected or
        unselected by mouse clicks.
    """

    # Pens used to draw the 'selection' marker:
    _selectedPenDark = wx.Pen(
        wx.SystemSettings_GetColour(wx.SYS_COLOUR_3DSHADOW), 1,
        wx.SOLID)

    _selectedPenLight = wx.Pen(
        wx.SystemSettings_GetColour(wx.SYS_COLOUR_3DHIGHLIGHT), 1,
        wx.SOLID)

    #-------------------------------------------------------------------------
    #  Initializes the object:
    #-------------------------------------------------------------------------

    def __init__(self, parent, bitmap, selected=None, handler=None,
                 padding=10):
        """ Initializes the object.
        """
        wx.Window.__init__(self, parent, -1,
                           size=wx.Size(bitmap.GetWidth() + padding,
                                        bitmap.GetHeight() + padding))
        self._bitmap = bitmap
        self._selected = selected
        self._handler = handler
        self._mouse_over = False
        self._button_down = False

        # Set up the 'paint' event handler:
        wx.EVT_PAINT(self, self._on_paint)

        # Set up mouse event handlers:
        wx.EVT_LEFT_DOWN(self, self._on_left_down)
        wx.EVT_LEFT_UP(self, self._on_left_up)
        wx.EVT_ENTER_WINDOW(self, self._on_enter)
        wx.EVT_LEAVE_WINDOW(self, self._on_leave)

    #-------------------------------------------------------------------------
    #  Gets/Sets the current selection state of the image:
    #-------------------------------------------------------------------------

    def Selected(self, selected=None):
        """ Gets or sets the selection state of the image.
        """
        if selected is not None:
            selected = (selected != 0)
            if selected != self._selected:
                if selected:
                    for control in self.GetParent().GetChildren():
                        if (isinstance(control, ImageControl) and
                                control.Selected()):
                            control.Selected(False)
                            break

                self._selected = selected
                self.Refresh()

        return self._selected

    #-------------------------------------------------------------------------
    #  Gets/Sets the current bitmap image:
    #-------------------------------------------------------------------------

    def Bitmap(self, bitmap=None):
        """ Gets or sets the bitmap image.
        """
        if bitmap is not None:
            if bitmap != self._bitmap:
                self._bitmap = bitmap
                self.Refresh()

        return self._bitmap

    #-------------------------------------------------------------------------
    #  Gets/Sets the current click handler:
    #-------------------------------------------------------------------------

    def Handler(self, handler=None):
        """ Gets or sets the click handler.
        """
        if handler is not None:
            if handler != self._handler:
                self._handler = handler
                self.Refresh()

        return self._handler

    #-------------------------------------------------------------------------
    #  Handles the mouse entering the control:
    #-------------------------------------------------------------------------

    def _on_enter(self, event=None):
        """ Handles the mouse entering the control.
        """
        if self._selected is not None:
            self._mouse_over = True
            self.Refresh()

    #-------------------------------------------------------------------------
    #  Handles the mouse leaving the control:
    #-------------------------------------------------------------------------

    def _on_leave(self, event=None):
        """ Handles the mouse leaving the control.
        """
        if self._mouse_over:
            self._mouse_over = False
            self.Refresh()

    #-------------------------------------------------------------------------
    #  Handles the user pressing the mouse button:
    #-------------------------------------------------------------------------

    def _on_left_down(self, event=None):
        """ Handles the user pressing the mouse button.
        """
        if self._selected is not None:
            self.CaptureMouse()
            self._button_down = True
            self.Refresh()

    #-------------------------------------------------------------------------
    #  Handles the user clicking the control:
    #-------------------------------------------------------------------------

    def _on_left_up(self, event=None):
        """ Handles the user clicking the control.
        """
        need_refresh = self._button_down
        if need_refresh:
            self.ReleaseMouse()
            self._button_down = False

        if self._selected is not None:
            wdx, wdy = self.GetClientSizeTuple()
            x = event.GetX()
            y = event.GetY()
            if (0 <= x < wdx) and (0 <= y < wdy):
                if self._selected != -1:
                    self.Selected(True)
                elif need_refresh:
                    self.Refresh()
                if self._handler is not None:
                    self._handler(self)
                return

        if need_refresh:
            self.Refresh()

    #-------------------------------------------------------------------------
    #  Handles the control being re-painted:
    #-------------------------------------------------------------------------

    def _on_paint(self, event=None):
        """ Handles the control being re-painted.
        """
        wdc = wx.PaintDC(self)
        wdx, wdy = self.GetClientSizeTuple()
        bitmap = self._bitmap
        bdx = bitmap.GetWidth()
        bdy = bitmap.GetHeight()
        wdc.DrawBitmap(bitmap, (wdx - bdx) / 2, (wdy - bdy) / 2, True)

        pens = [self._selectedPenLight, self._selectedPenDark]
        bd = self._button_down

        if self._mouse_over:
            wdc.SetBrush(wx.TRANSPARENT_BRUSH)
            wdc.SetPen(pens[bd])
            wdc.DrawLine(0, 0, wdx, 0)
            wdc.DrawLine(0, 1, 0, wdy)
            wdc.SetPen(pens[1 - bd])
            wdc.DrawLine(wdx - 1, 1, wdx - 1, wdy)
            wdc.DrawLine(1, wdy - 1, wdx - 1, wdy - 1)

        if self._selected is True:
            wdc.SetBrush(wx.TRANSPARENT_BRUSH)
            wdc.SetPen(pens[bd])
            wdc.DrawLine(1, 1, wdx - 1, 1)
            wdc.DrawLine(1, 1, 1, wdy - 1)
            wdc.DrawLine(2, 2, wdx - 2, 2)
            wdc.DrawLine(2, 2, 2, wdy - 2)
            wdc.SetPen(pens[1 - bd])
            wdc.DrawLine(wdx - 2, 2, wdx - 2, wdy - 1)
            wdc.DrawLine(2, wdy - 2, wdx - 2, wdy - 2)
            wdc.DrawLine(wdx - 3, 3, wdx - 3, wdy - 2)
            wdc.DrawLine(3, wdy - 3, wdx - 3, wdy - 3)



#------------------------------------------------------------------------------
#
#  Copyright (c) 2008, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/21/2004
#
#------------------------------------------------------------------------------

""" Defines the instance editor factory for all traits user interface
toolkits.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import Str, List, Enum, Unicode, Type, Bool

from ..view import View, AKind

from ..ui_traits import AView

from ..instance_choice import InstanceChoice, InstanceChoiceItem

from ..editor_factory import EditorFactory

#-------------------------------------------------------------------------
#  'ToolkitEditorFactory' class:
#-------------------------------------------------------------------------


class ToolkitEditorFactory(EditorFactory):
    """ Editor factory for instance editors.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # List of items describing the types of selectable or editable instances
    values = List(InstanceChoiceItem)

    # Extended name of the context object trait containing the list of types of
    # selectable or editable instances
    name = Str

    # Is the current value of the object trait editable (vs. merely
    # selectable)?
    editable = Bool(True)

    # Should the object trait value be selectable from a list of objects (a
    # value of True forces a selection list to be displayed, while a value of
    # False displays a selection list only if at least one object in the list
    # of possible object values is selectable):
    selectable = Bool(False)

    # Should the editor support drag and drop of objects to set the trait value
    # (a value of True forces the editor to allow drag and drop, while a value
    # of False only supports drag and drop if at least one item in the list of
    # possible objects supports drag and drop):
    droppable = Bool(False)

    # Should factory-created objects be cached?
    cachable = Bool(True)

    # Optional label for button
    label = Unicode

    # Optional instance view to use
    view = AView

    # Extended name of the context object trait containing the view, or name of
    # the view, to use
    view_name = Str

    # The ID to use with the view
    id = Str

    # Kind of pop-up editor (live, modal, nonmodal, wizard)
    kind = AKind

    # The orientation of the instance editor relative to the instance selector
    orientation = Enum('default', 'horizontal', 'vertical')

    # The default adapter class used to create InstanceChoice compatible
    # adapters for instance objects:
    adapter = Type(InstanceChoice, allow_none=False)

    #-------------------------------------------------------------------------
    #  Traits view definitions:
    #-------------------------------------------------------------------------

    traits_view = View([['label{Button label}',
                         'view{View name}', '|[]'],
                        ['kind@', '|[Pop-up editor style]<>']])


# Define the InstanceEditor class.
InstanceEditor = ToolkitEditorFactory

### EOF #######################################################################



from svg_parser_utils import *
import OpenGL.GL as gl
import math

EPSILON = 0.001


class vec2(object):
    def __init__(self, *args):
        if isinstance(args[0], vec2):
            self.x = args[0].x
            self.y = args[0].y
        elif isinstance(args[0], list):
            self.x, self.y = args[0]
        else:
            self.x, self.y = args[0], args[1]

    def tolist(self):
        return [self.x, self.y]

    def __repr__(self):
        return '(' + str(self.x) + ',' + str(self.y) + ')'

    def __neg__(self):
        return vec2(-self.x, -self.y)

    def __abs__(self):
        return self.length()

    def __add__(self, other):
        return vec2(self.x + other.x, self.y + other.y)

    def __sub__(self, other):
        return vec2(self.x - other.x, self.y - other.y)

    def __mul__(self, scale):
        return vec2(self.x * scale, self.y * scale)

    def __div__(self, scale):
        return vec2(self.x / scale, self.y / scale)

    def __eq__(self, other):
        if not other: return False
        return abs(self.x - other.x) < EPSILON and abs(self.y - other.y) < EPSILON

    def __ne__(self, other):
        return not(self.__eq__(other))

    def normalized(self):
        l = self.length()
        if l == 0:
            return vec2(1, 0)
        else:
            return vec2(self.x, self.y) / self.length()

    def angle(self):
        return math.atan2(self.y, self.x)

    def length(self):
        return math.sqrt(self.x ** 2 + self.y ** 2)


def intersection(p1, p2, p3, p4):
    """
    Returns whether two lines intersected, and the point at which
    they intersected (or, "None" if the lines are parallel)
    """
    A1 = p2.y - p1.y
    B1 = p1.x - p2.x
    C1 = A1 * p1.x + B1 * p1.y

    A2 = p4.y - p3.y
    B2 = p3.x - p4.x
    C2 = A2 * p3.x + B2 * p3.y

    det = A1 * B2 - A2 * B1

    if abs(det) < EPSILON:  # Lines are parallel
        if (p1 == p3) or (p1 == p4):
            return True, p1
        elif (p2 == p3) or (p2 == p4):
            return True, p2
        return False, None
    else:
        result = vec2(
            (B2 * C1 - B1 * C2) / det,
            (A1 * C2 - A2 * C1) / det)

    epsilon = .01

    on_line_segment = True
    on_line_segment &= result.x >= (min(p1.x, p2.x) - epsilon)
    on_line_segment &= result.y >= (min(p1.y, p2.y) - epsilon)
    on_line_segment &= result.x <= (max(p1.x, p2.x) + epsilon)
    on_line_segment &= result.y <= (max(p1.y, p2.y) + epsilon)

    return on_line_segment, result


def line_length(a, b):
    return math.sqrt((b.x - a.x) ** 2 + (b.y - a.y) ** 2)


def radian(deg):
    return deg * (math.pi / 180.0)

ninety_degrees = radian(90)


class Matrix(object):
    def __init__(self, string=None):
        self.values = [1, 0, 0, 1, 0, 0]
        if isinstance(string, str):
            string = string.strip()
            if string.startswith('matrix('):
                self.values = [float(x) for x in parse_list(string[7:-1])]
            elif string.startswith('translate('):
                args = [float(x) for x in parse_list(string[10:-1])]

                #if len(args) == 2:
                self.values = [1, 0, 0, 1, args[0], args[1]]
                #else:
                #    self.values = [1, 0, 0, 1, args[0], 0]
            elif string.startswith('scale('):
                inside = string[6:-1]
                scale_vars = [float(x) for x in parse_float_list(inside)]

                if len(scale_vars) == 1:
                    self.values = [scale_vars[0], 0, 0, scale_vars[0], 0, 0]
                else:
                    self.values = [scale_vars[0], 0, 0, scale_vars[1], 0, 0]
            elif string.startswith('rotate('):
                angle = float(string[7:-1])
                theta = radian(angle)

                #where does angle go in here?
                self.values = [math.cos(theta), math.sin(theta), -math.sin(theta), math.cos(theta), 0, 0]
        elif string is not None:
            self.values = list(string)

    def __enter__(self):
        gl.glPushMatrix()
        gl.glMultMatrixf(self.to_mat4())
        return self

    def __exit__(self, type, value, traceback):
        gl.glPopMatrix()

    def __call__(self, other):
        return (self.values[0] * other[0] + self.values[2] * other[1] + self.values[4],
                self.values[1] * other[0] + self.values[3] * other[1] + self.values[5])
    
    def __str__(self):
        return str(self.values)
    
    def to_mat4(self):
        v = self.values
        return [v[0], v[1], 0.0, 0.0,
                v[2], v[3], 0.0, 0.0,
                0.0,  0.0,  1.0, 0.0,
                v[4], v[5], 0.0, 1.0]

    @classmethod
    def identity(cls):
        return Matrix([1, 0, 0, 1, 0, 0])

    @classmethod
    def translation(cls, x, y):
        return Matrix([1, 0, 0, 1, x, y])

    @classmethod
    def scale(cls, sx, sy):
        return Matrix([sx, 0, 0, sy, 0, 0])

    @classmethod
    def rotation(cls, theta):
        return Matrix([math.cos(theta), math.sin(theta), -math.sin(theta), math.cos(theta), 0, 0])

    @classmethod
    def transform(cls, x, y, theta=0, sx=1, sy=1):
        return Matrix([math.cos(theta) * sx, math.sin(theta),
                       -math.sin(theta), math.cos(theta) * sy,
                       x, y])

    def inverse(self):
        d = float(self.values[0] * self.values[3] - self.values[1] * self.values[2])
        return Matrix([self.values[3] / d, -self.values[1] / d, -self.values[2] / d, self.values[0] / d,
                       (self.values[2] * self.values[5] - self.values[3] * self.values[4]) / d,
                       (self.values[1] * self.values[4] - self.values[0] * self.values[5]) / d])

    def __mul__(self, other):
        a, b, c, d, e, f = self.values
        u, v, w, x, y, z = other.values
        return Matrix([
            a * u + c * v,
            b * u + d * v,
            a * w + c * x,
            b * w + d * x,
            a * y + c * z + e,
            b * y + d * z + f])


def svg_matrix_to_gl_matrix(matrix):
    v = matrix.values
    return [v[0], v[1], 0.0, v[2], v[3], 0.0, v[4], v[5], 1.0]


class BoundingBox:

    def __init__(self, point_cloud=None):
        self.min_x = None
        self.max_x = None
        self.min_y = None
        self.max_y = None
        if point_cloud:
            self.expand(point_cloud)

    def expand(self, points):
        for p in points:
            x, y = p[0], p[1]
            if self.min_x is None or x < self.min_x:
                self.min_x = x
            if self.min_y is None or y < self.min_y:
                self.min_y = y
            if self.max_x is None or x > self.max_x:
                self.max_x = x
            if self.max_y is None or y > self.max_y:
                self.max_y = y

    def extents(self):
        return self.min_x, self.min_y, self.max_x, self.max_y


#  Copyright (c) 2009-2010, Cloud Matrix Pty. Ltd.
#  All rights reserved; available under the terms of the MIT License.
"""

  fs.contrib.davfs.xmlobj:  dexml model definitions for WebDAV

This module defines the various XML element structures for WebDAV as a set
of dexml.Model subclasses.

"""

from urlparse import urlparse, urlunparse

from httplib import responses as STATUS_CODE_TEXT
STATUS_CODE_TEXT[207] = "Multi-Status"

import dexml
from dexml import fields

Error = dexml.Error


class _davbase(dexml.Model):
    """Base class for all davfs XML models."""

    class meta:
        namespace = "DAV:"
        namespace_prefix = "D"
        order_sensitive = False


class HrefField(fields.String):
    """Field representing a <href> tag."""

    def __init__(self,*args,**kwds):
        kwds["tagname"] = "href"
        super(HrefField,self).__init__(*args,**kwds)

    def parse_value(self,value):
        url = urlparse(value.encode("UTF-8"))
        return urlunparse((url.scheme,url.netloc,url.path,url.params,url.query,url.fragment))

    def render_value(self,value):
        url = urlparse(value.encode("UTF-8"))
        return urlunparse((url.scheme,url.netloc,url.path,url.params,url.query,url.fragment))


class TimeoutField(fields.Field):
    """Field representing a WebDAV timeout value."""

    def __init__(self,*args,**kwds):
        if "tagname" not in kwds:
            kwds["tagname"] = "timeout"
        super(TimeoutField,self).__init__(*args,**kwds)

    @classmethod
    def parse_value(cls,value):
        if value == "Infinite":
            return None
        if value.startswith("Second-"):
            return int(value[len("Second-"):])
        raise ValueError("invalid timeout specifier: %s" % (value,))

    def render_value(self,value):
        if value is None:
            return "Infinite"
        else:
            return "Second-" + str(value)


class StatusField(fields.Value):
    """Field representing a WebDAV status-line value.

    The value may be set as either a string or an integer, and is converted
    into a StatusString instance.
    """

    def __init__(self,*args,**kwds):
        kwds["tagname"] = "status"
        super(StatusField,self).__init__(*args,**kwds)

    def __get__(self,instance,owner):
        val = super(StatusField,self).__get__(instance,owner)
        if val is not None:
            val = StatusString(val,instance,self)
        return val

    def __set__(self,instance,value):
        if isinstance(value,basestring):
            # sanity check it
            bits = value.split(" ")
            if len(bits) < 3 or bits[0] != "HTTP/1.1":
                raise ValueError("Not a valid status: '%s'" % (value,))
            int(bits[1])
        elif isinstance(value,int):
            # convert it to a message
            value = StatusString._value_for_code(value)
        super(StatusField,self).__set__(instance,value)


class StatusString(str):
    """Special string representing a HTTP status line.

    It's a string, but it exposes the integer attribute "code" giving just
    the actual response code.
    """

    def __new__(cls,val,inst,owner):
        return str.__new__(cls,val)

    def __init__(self,val,inst,owner):
         self._owner = owner
         self._inst = inst

    @staticmethod
    def _value_for_code(code):
        msg = STATUS_CODE_TEXT.get(code,"UNKNOWN STATUS CODE")
        return "HTTP/1.1 %d %s" % (code,msg)

    def _get_code(self):
        return int(self.split(" ")[1])
    def _set_code(self,code):
        newval = self._value_for_code(code)
        self._owner.__set__(self._inst,newval)
    code = property(_get_code,_set_code)


class multistatus(_davbase):
    """XML model for a multi-status response message."""
    responses = fields.List("response",minlength=1)
    description = fields.String(tagname="responsedescription",required=False)


class response(_davbase):
    """XML model for an individual response in a multi-status message."""
    href = HrefField()
    # TODO: ensure only one of hrefs/propstats
    hrefs = fields.List(HrefField(),required=False)
    status = StatusField(required=False)
    propstats = fields.List("propstat",required=False)
    description = fields.String(tagname="responsedescription",required=False)


class propstat(_davbase):
    """XML model for a propstat response message."""
    props = fields.XmlNode(tagname="prop",encoding="UTF-8")
    status = StatusField()
    description = fields.String(tagname="responsedescription",required=False)


class propfind(_davbase):
    """XML model for a propfind request message."""
    allprop = fields.Boolean(tagname="allprop",required=False)
    propname = fields.Boolean(tagname="propname",required=False)
    prop = fields.XmlNode(tagname="prop",required=False,encoding="UTF-8")


class propertyupdate(_davbase):
    """XML model for a propertyupdate request message."""
    commands = fields.List(fields.Choice("remove","set"))

class remove(_davbase):
    """XML model for a propertyupdate remove command."""
    props = fields.XmlNode(tagname="prop",encoding="UTF-8")

class set(_davbase):
    """XML model for a propertyupdate set command."""
    props = fields.XmlNode(tagname="prop",encoding="UTF-8")

class lockdiscovery(_davbase):
    """XML model for a lockdiscovery request message."""
    locks = fields.List("activelock")

class activelock(_davbase):
    """XML model for an activelock response message."""
    lockscope = fields.Model("lockscope")
    locktype = fields.Model("locktype")
    depth = fields.String(tagname="depth")
    owner = fields.XmlNode(tagname="owner",encoding="UTF-8",required=False)
    timeout = TimeoutField(required=False)
    locktoken = fields.Model("locktoken",required=False)

class lockscope(_davbase):
    """XML model for a lockscope response message."""
    shared = fields.Boolean(tagname="shared",empty_only=True)
    exclusive = fields.Boolean(tagname="exclusive",empty_only=True)

class locktoken(_davbase):
    """XML model for a locktoken response message."""
    tokens = fields.List(HrefField())

class lockentry(_davbase):
    """XML model for a lockentry response message."""
    lockscope = fields.Model("lockscope")
    locktype = fields.Model("locktype")

class lockinfo(_davbase):
    """XML model for a lockinfo response message."""
    lockscope = fields.Model("lockscope")
    locktype = fields.Model("locktype")
    owner = fields.XmlNode(tagname="owner",encoding="UTF-8")

class locktype(_davbase):
    """XML model for a locktype response message."""
    type = fields.XmlNode(encoding="UTF-8")





#-------------------------------------------------------------------------
#
#  Copyright (c) 2007, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   07/13/2007
#
#-------------------------------------------------------------------------

""" Defines 'theme' related classes.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import HasPrivateTraits

from .ui_traits import Image, HasBorder, HasMargin, Alignment

#-------------------------------------------------------------------------
#  'Theme' class:
#-------------------------------------------------------------------------


class Theme(HasPrivateTraits):

    #-- Public Traits --------------------------------------------------------

    # The background image to use for the theme:
    image = Image

    # The border inset:
    border = HasBorder

    # The margin to use around the content:
    content = HasMargin

    # The margin to use around the label:
    label = HasMargin

    # The alignment to use for positioning the label:
    alignment = Alignment(cols=4)

    # Note: The 'content_color' and 'label_color' traits should be added by a
    # toolkit-specific category...

    #-- Constructor ----------------------------------------------------------

    def __init__(self, image=None, **traits):
        """ Initializes the object.
        """
        if image is not None:
            self.image = image

        super(Theme, self).__init__(**traits)

# Create a default theme:
default_theme = Theme()



__author__ = 'Ian'
import OpenGL.GL as gl
import graphics


class Texture2D:

    def __init__(self, w, h, wrap=True):

        self.width = w
        self.height = h
        self.id = gl.glGenTextures(1)
        print "texture id", self.id

        self.bind()
        gl.glTexEnvf(gl.GL_TEXTURE_ENV, gl.GL_TEXTURE_ENV_MODE, gl.GL_MODULATE)

        gl.glTexParameterf(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER, gl.GL_NEAREST)
        gl.glTexParameterf(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER, gl.GL_NEAREST)

        wrap_mode = gl.GL_REPEAT if wrap else gl.GL_CLAMP
        gl.glTexParameterf(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_S, wrap_mode)
        gl.glTexParameterf(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_T, wrap_mode)

        gl.glTexImage2D(gl.GL_TEXTURE_2D, 0, gl.GL_RGBA8, self.width, self.height, 0, gl.GL_RGBA, gl.GL_UNSIGNED_BYTE, None)

        print "Tex OK? ", gl.glGetError() == gl.GL_NO_ERROR
        self.unbind()

    def resize(self, w, h):
        self.width = w
        self.height = h
        self.bind()
        gl.glTexImage2D(gl.GL_TEXTURE_2D, 0, gl.GL_RGBA8, self.width, self.height, 0, gl.GL_RGBA, gl.GL_UNSIGNED_BYTE, None)
        self.unbind()


    def bind(self):
        gl.glBindTexture(gl.GL_TEXTURE_2D, self.id)

    def unbind(self):
        gl.glBindTexture(gl.GL_TEXTURE_2D, 0)

    def __enter__(self):
        self.bind()
        return self

    def __exit__(self, type, value, traceback):
        self.unbind()


class RenderBufferObject:
    def __init__(self, w, h):
        self.id = gl.glGenRenderbuffers(1)
        self.resize(w, h)

    def resize(self, w, h):
        self.width, self.height = w, h
        self.bind()
        gl.glRenderbufferStorage(gl.GL_RENDERBUFFER, gl.GL_DEPTH24_STENCIL8, w, h)
        self.unbind()

    def bind(self):
        gl.glBindRenderbuffer(gl.GL_RENDERBUFFER, self.id)

    def unbind(self):
        gl.glBindRenderbuffer(gl.GL_RENDERBUFFER, 0)


class RenderTarget:

    id_stack = []

    def __init__(self, w, h, depth_and_stencil=False):
        self.texture = Texture2D(w, h)
        self.id = gl.glGenFramebuffers(1)
        self.bind()
        self.depth_stencil = None
        if depth_and_stencil:
            self.depth_stencil = RenderBufferObject(w, h)

        gl.glFramebufferTexture2D(
            gl.GL_FRAMEBUFFER,
            gl.GL_COLOR_ATTACHMENT0,
            gl.GL_TEXTURE_2D,
            self.texture.id,
            0)

        self.ok = self.check_status()

        self.unbind()

    def check_status(self):
        status = gl.glCheckFramebufferStatus(gl.GL_FRAMEBUFFER)
        if status == gl.GL_FRAMEBUFFER_COMPLETE:
            print "Framebuffer complete"
            return True
        else:
            print "Render target error: " + str(status)
            return False

    def blit(self):
        w,h = self.texture.width, self.texture.height
        with self.texture:
            graphics.draw_quad(0.5, 0.5, w+0.5, h+0.5)

    def resize(self, w, h):
        self.texture.resize(w, h)
        if self.depth_stencil:
            self.depth_stencil.resize(w, h)

    def bind(self):
        gl.glBindFramebuffer(gl.GL_FRAMEBUFFER, 0)
        gl.glBindFramebuffer(gl.GL_FRAMEBUFFER, self.id)
        RenderTarget.id_stack.append(self.id)

    def unbind(self):
        if len(RenderTarget.id_stack):
            del RenderTarget.id_stack[-1]

        if len(RenderTarget.id_stack) >= 1:
            id = RenderTarget.id_stack[-1]
            gl.glBindFramebuffer(gl.GL_FRAMEBUFFER, 0)
            gl.glBindFramebuffer(gl.GL_FRAMEBUFFER, id)
        else:
            gl.glBindFramebuffer(gl.GL_FRAMEBUFFER, 0)

    def __enter__(self):
        self.bind()
        return self

    def __exit__(self, type, value, traceback):
        self.unbind()
        pass


class CanvasManager:
    instance = None

    def __init__(self):
        print "Initializing canvas manager"
        self.canvas = {}
        vp = list(gl.glGetFloatv(gl.GL_VIEWPORT))
        self.w = int(vp[2])
        self.h = int(vp[3])

    def resize(self, w, h):
        print "resizing", w, h
        self.w = w
        self.h = h

        for c in self.canvas.values():
            c.resize(w, h)

    def get(self, name):
        if not name in self.canvas:
            self.canvas[name] = RenderTarget(self.w, self.h)
        return self.canvas[name]

    def temp(self):
        return self.get('temp' + str(len(RenderTarget.id_stack)))

    def update(self):
        vp = list(gl.glGetFloatv(gl.GL_VIEWPORT))

        if self.w != int(vp[2]) and self.h != int(vp[3]):
            self.resize(int(vp[2]), int(vp[3]))

    @classmethod
    def inst(cls):
        if not cls.instance:
            cls.instance = CanvasManager()
        return cls.instance


"""
fs.wrapfs
=========

A class for wrapping an existing FS object with additional functionality.

This module provides the class WrapFS, a base class for objects that wrap
another FS object and provide some transformation of its contents.  It could
be very useful for implementing e.g. transparent encryption or compression
services.

For a simple example of how this class could be used, see the 'HideDotFilesFS'
class in the module fs.wrapfs.hidedotfilesfs.  This wrapper implements the
standard unix shell functionality of hiding dot-files in directory listings.

"""

import re
import sys
import fnmatch
import threading

from fs.base import FS, threading, synchronize, NoDefaultMeta
from fs.errors import *
from fs.path import *
from fs.local_functools import wraps


def rewrite_errors(func):
    """Re-write paths in errors raised by wrapped FS objects."""
    @wraps(func)
    def wrapper(self,*args,**kwds):
        try:
            return func(self,*args,**kwds)
        except ResourceError, e:
            (exc_type,exc_inst,tb) = sys.exc_info()
            try:
                e.path = self._decode(e.path)
            except (AttributeError, ValueError, TypeError):
                raise e, None, tb
            raise
    return wrapper


class WrapFS(FS):
    """FS that wraps another FS, providing translation etc.

    This class allows simple transforms to be applied to the names
    and/or contents of files in an FS.  It could be used to implement
    e.g. compression or encryption in a relatively painless manner.

    The following methods can be overridden to control how files are
    accessed in the underlying FS object:

     * _file_wrap(file, mode):  called for each file that is opened from
                                the underlying FS; may return a modified
                                file-like object.

     *  _encode(path):  encode a path for access in the underlying FS

     *  _decode(path):  decode a path from the underlying FS

    If the required path translation proceeds one component at a time,
    it may be simpler to override the _encode_name() and _decode_name()
    methods.
    """

    def __init__(self, fs):
        super(WrapFS, self).__init__()
        try:
            self._lock = fs._lock
        except (AttributeError,FSError):
            self._lock = self._lock = threading.RLock()
        self.wrapped_fs = fs

    def _file_wrap(self, f, mode):
        """Apply wrapping to an opened file."""
        return f

    def _encode_name(self, name):
        """Encode path component for the underlying FS."""
        return name

    def _decode_name(self, name):
        """Decode path component from the underlying FS."""
        return name

    def _encode(self, path):
        """Encode path for the underlying FS."""
        e_names = []
        for name in iteratepath(path):
            if name == "":
                e_names.append("")
            else:
                e_names.append(self._encode_name(name))
        return "/".join(e_names)

    def _decode(self, path):
        """Decode path from the underlying FS."""
        d_names = []
        for name in iteratepath(path):
            if name == "":
                d_names.append("")
            else:
                d_names.append(self._decode_name(name))
        return "/".join(d_names)

    def _adjust_mode(self, mode):
        """Adjust the mode used to open a file in the underlying FS.

        This method takes the mode given when opening a file, and should
        return a two-tuple giving the mode to be used in this FS as first
        item, and the mode to be used in the underlying FS as the second.

        An example of why this is needed is a WrapFS subclass that does
        transparent file compression - in this case files from the wrapped
        FS cannot be opened in append mode.
        """
        return (mode, mode)

    def __unicode__(self):
        return u"<%s: %s>" % (self.__class__.__name__,self.wrapped_fs,)

    #def __str__(self):
    #    return unicode(self).encode(sys.getdefaultencoding(),"replace")


    @rewrite_errors
    def getmeta(self, meta_name, default=NoDefaultMeta):
        return self.wrapped_fs.getmeta(meta_name, default)

    @rewrite_errors
    def hasmeta(self, meta_name):
        return self.wrapped_fs.hasmeta(meta_name)

    @rewrite_errors
    def validatepath(self, path):
        return self.wrapped_fs.validatepath(self._encode(path))

    @rewrite_errors
    def getsyspath(self, path, allow_none=False):
        return self.wrapped_fs.getsyspath(self._encode(path), allow_none)

    @rewrite_errors
    def getpathurl(self, path, allow_none=False):
        return self.wrapped_fs.getpathurl(self._encode(path), allow_none)

    @rewrite_errors
    def hassyspath(self, path):
        return self.wrapped_fs.hassyspath(self._encode(path))

    @rewrite_errors
    def open(self, path, mode='r', **kwargs):
        (mode, wmode) = self._adjust_mode(mode)
        f = self.wrapped_fs.open(self._encode(path), wmode, **kwargs)
        return self._file_wrap(f, mode)

    @rewrite_errors
    def setcontents(self, path, data, encoding=None, errors=None, chunk_size=64*1024):
        #  We can't pass setcontents() through to the wrapped FS if the
        #  wrapper has defined a _file_wrap method, as it would bypass
        #  the file contents wrapping.
        #if self._file_wrap.im_func is WrapFS._file_wrap.im_func:
        if getattr(self.__class__, '_file_wrap', None) is getattr(WrapFS, '_file_wrap', None):
            return self.wrapped_fs.setcontents(self._encode(path), data, encoding=encoding, errors=errors, chunk_size=chunk_size)
        else:
            return super(WrapFS, self).setcontents(path, data, encoding=encoding, errors=errors, chunk_size=chunk_size)

    @rewrite_errors
    def createfile(self, path, wipe=False):
        return self.wrapped_fs.createfile(self._encode(path), wipe=wipe)

    @rewrite_errors
    def exists(self, path):
        return self.wrapped_fs.exists(self._encode(path))

    @rewrite_errors
    def isdir(self, path):
        return self.wrapped_fs.isdir(self._encode(path))

    @rewrite_errors
    def isfile(self, path):
        return self.wrapped_fs.isfile(self._encode(path))

    @rewrite_errors
    def listdir(self, path="", wildcard=None, full=False, absolute=False, dirs_only=False, files_only=False):
        kwds = dict(wildcard=wildcard,
                    full=full,
                    absolute=absolute,
                    dirs_only=dirs_only,
                    files_only=files_only)
        full = kwds.pop("full",False)
        absolute = kwds.pop("absolute",False)
        wildcard = kwds.pop("wildcard",None)
        if wildcard is None:
            wildcard = lambda fn:True
        elif not callable(wildcard):
            wildcard_re = re.compile(fnmatch.translate(wildcard))
            wildcard = lambda fn:bool (wildcard_re.match(fn))
        entries = []
        enc_path = self._encode(path)
        for e in self.wrapped_fs.listdir(enc_path,**kwds):
            e = basename(self._decode(pathcombine(enc_path,e)))
            if not wildcard(e):
                continue
            if full:
                e = pathcombine(path,e)
            elif absolute:
                e = abspath(pathcombine(path,e))
            entries.append(e)
        return entries

    @rewrite_errors
    def ilistdir(self, path="", wildcard=None, full=False, absolute=False, dirs_only=False, files_only=False):
        kwds = dict(wildcard=wildcard,
                    full=full,
                    absolute=absolute,
                    dirs_only=dirs_only,
                    files_only=files_only)
        full = kwds.pop("full",False)
        absolute = kwds.pop("absolute",False)
        wildcard = kwds.pop("wildcard",None)
        if wildcard is None:
            wildcard = lambda fn:True
        elif not callable(wildcard):
            wildcard_re = re.compile(fnmatch.translate(wildcard))
            wildcard = lambda fn:bool (wildcard_re.match(fn))
        enc_path = self._encode(path)
        for e in self.wrapped_fs.ilistdir(enc_path,**kwds):
            e = basename(self._decode(pathcombine(enc_path,e)))
            if not wildcard(e):
                continue
            if full:
                e = pathcombine(path,e)
            elif absolute:
                e = abspath(pathcombine(path,e))
            yield e

    @rewrite_errors
    def listdirinfo(self, path="", wildcard=None, full=False, absolute=False, dirs_only=False, files_only=False):
        kwds = dict(wildcard=wildcard,
                    full=full,
                    absolute=absolute,
                    dirs_only=dirs_only,
                    files_only=files_only)
        full = kwds.pop("full",False)
        absolute = kwds.pop("absolute",False)
        wildcard = kwds.pop("wildcard",None)
        if wildcard is None:
            wildcard = lambda fn:True
        elif not callable(wildcard):
            wildcard_re = re.compile(fnmatch.translate(wildcard))
            wildcard = lambda fn:bool (wildcard_re.match(fn))
        entries = []
        enc_path = self._encode(path)
        for (nm,info) in self.wrapped_fs.listdirinfo(enc_path,**kwds):
            nm = basename(self._decode(pathcombine(enc_path,nm)))
            if not wildcard(nm):
                continue
            if full:
                nm = pathcombine(path,nm)
            elif absolute:
                nm = abspath(pathcombine(path,nm))
            entries.append((nm,info))
        return entries

    @rewrite_errors
    def ilistdirinfo(self, path="", wildcard=None, full=False, absolute=False, dirs_only=False, files_only=False):
        kwds = dict(wildcard=wildcard,
                    full=full,
                    absolute=absolute,
                    dirs_only=dirs_only,
                    files_only=files_only)
        full = kwds.pop("full",False)
        absolute = kwds.pop("absolute",False)
        wildcard = kwds.pop("wildcard",None)
        if wildcard is None:
            wildcard = lambda fn:True
        elif not callable(wildcard):
            wildcard_re = re.compile(fnmatch.translate(wildcard))
            wildcard = lambda fn:bool (wildcard_re.match(fn))
        enc_path = self._encode(path)
        for (nm,info) in self.wrapped_fs.ilistdirinfo(enc_path,**kwds):
            nm = basename(self._decode(pathcombine(enc_path,nm)))
            if not wildcard(nm):
                continue
            if full:
                nm = pathcombine(path,nm)
            elif absolute:
                nm = abspath(pathcombine(path,nm))
            yield (nm,info)

    @rewrite_errors
    def walk(self,path="/",wildcard=None,dir_wildcard=None,search="breadth",ignore_errors=False):
        if dir_wildcard is not None:
            #  If there is a dir_wildcard, fall back to the default impl
            #  that uses listdir().  Otherwise we run the risk of enumerating
            #  lots of directories that will just be thrown away.
            for item in super(WrapFS,self).walk(path,wildcard,dir_wildcard,search,ignore_errors):
                yield item
        #  Otherwise, the wrapped FS may provide a more efficient impl
        #  which we can use directly.
        else:
            if wildcard is not None and not callable(wildcard):
                wildcard_re = re.compile(fnmatch.translate(wildcard))
                wildcard = lambda fn:bool (wildcard_re.match(fn))
            for (dirpath,filepaths) in self.wrapped_fs.walk(self._encode(path),search=search,ignore_errors=ignore_errors):
                filepaths = [basename(self._decode(pathcombine(dirpath,p)))
                                 for p in filepaths]
                dirpath = abspath(self._decode(dirpath))
                if wildcard is not None:
                    filepaths = [p for p in filepaths if wildcard(p)]
                yield (dirpath,filepaths)

    @rewrite_errors
    def walkfiles(self,path="/",wildcard=None,dir_wildcard=None,search="breadth",ignore_errors=False):
        if dir_wildcard is not None:
            #  If there is a dir_wildcard, fall back to the default impl
            #  that uses listdir().  Otherwise we run the risk of enumerating
            #  lots of directories that will just be thrown away.
            for item in super(WrapFS,self).walkfiles(path,wildcard,dir_wildcard,search,ignore_errors):
                yield item
        #  Otherwise, the wrapped FS may provide a more efficient impl
        #  which we can use directly.
        else:
            if wildcard is not None and not callable(wildcard):
                wildcard_re = re.compile(fnmatch.translate(wildcard))
                wildcard = lambda fn:bool (wildcard_re.match(fn))
            for filepath in self.wrapped_fs.walkfiles(self._encode(path),search=search,ignore_errors=ignore_errors):
                filepath = abspath(self._decode(filepath))
                if wildcard is not None:
                    if not wildcard(basename(filepath)):
                        continue
                yield filepath

    @rewrite_errors
    def walkdirs(self,path="/",wildcard=None,search="breadth",ignore_errors=False):
        if wildcard is not None:
            #  If there is a wildcard, fall back to the default impl
            #  that uses listdir().  Otherwise we run the risk of enumerating
            #  lots of directories that will just be thrown away.
            for item in super(WrapFS,self).walkdirs(path,wildcard,search,ignore_errors):
                yield item
        #  Otherwise, the wrapped FS may provide a more efficient impl
        #  which we can use directly.
        else:
            for dirpath in self.wrapped_fs.walkdirs(self._encode(path),search=search,ignore_errors=ignore_errors):
                yield abspath(self._decode(dirpath))


    @rewrite_errors
    def makedir(self, path, *args, **kwds):
        return self.wrapped_fs.makedir(self._encode(path),*args,**kwds)

    @rewrite_errors
    def remove(self, path):
        return self.wrapped_fs.remove(self._encode(path))

    @rewrite_errors
    def removedir(self, path, *args, **kwds):
        return self.wrapped_fs.removedir(self._encode(path),*args,**kwds)

    @rewrite_errors
    def rename(self, src, dst):
        return self.wrapped_fs.rename(self._encode(src),self._encode(dst))

    @rewrite_errors
    def getinfo(self, path):
        return self.wrapped_fs.getinfo(self._encode(path))

    @rewrite_errors
    def settimes(self, path, *args, **kwds):
        return self.wrapped_fs.settimes(self._encode(path), *args,**kwds)

    @rewrite_errors
    def desc(self, path):
        return self.wrapped_fs.desc(self._encode(path))

    @rewrite_errors
    def copy(self, src, dst, **kwds):
        return self.wrapped_fs.copy(self._encode(src),self._encode(dst),**kwds)

    @rewrite_errors
    def move(self, src, dst, **kwds):
        return self.wrapped_fs.move(self._encode(src),self._encode(dst),**kwds)

    @rewrite_errors
    def movedir(self, src, dst, **kwds):
        return self.wrapped_fs.movedir(self._encode(src),self._encode(dst),**kwds)

    @rewrite_errors
    def copydir(self, src, dst, **kwds):
        return self.wrapped_fs.copydir(self._encode(src),self._encode(dst),**kwds)

    @rewrite_errors
    def getxattr(self, path, name, default=None):
        try:
            return self.wrapped_fs.getxattr(self._encode(path),name,default)
        except AttributeError:
            raise UnsupportedError("getxattr")

    @rewrite_errors
    def setxattr(self, path, name, value):
        try:
            return self.wrapped_fs.setxattr(self._encode(path),name,value)
        except AttributeError:
            raise UnsupportedError("setxattr")

    @rewrite_errors
    def delxattr(self, path, name):
        try:
            return self.wrapped_fs.delxattr(self._encode(path),name)
        except AttributeError:
            raise UnsupportedError("delxattr")

    @rewrite_errors
    def listxattrs(self, path):
        try:
            return self.wrapped_fs.listxattrs(self._encode(path))
        except AttributeError:
            raise UnsupportedError("listxattrs")

    def __getattr__(self, attr):
        #  These attributes can be used by the destructor, but may not be
        #  defined if there are errors in the constructor.
        if attr == "closed":
            return False
        if attr == "wrapped_fs":
            return None
        if attr.startswith("_"):
            raise AttributeError(attr)
        return getattr(self.wrapped_fs,attr)

    @rewrite_errors
    def close(self):
        if not self.closed:
            self.wrapped_fs.close()
            super(WrapFS,self).close()
            self.wrapped_fs = None


def wrap_fs_methods(decorator, cls=None, exclude=[]):
    """Apply the given decorator to all FS methods on the given class.

    This function can be used in two ways.  When called with two arguments it
    applies the given function 'decorator' to each FS method of the given
    class.  When called with just a single argument, it creates and returns
    a class decorator which will do the same thing when applied.  So you can
    use it like this::

        wrap_fs_methods(mydecorator,MyFSClass)

    Or on more recent Python versions, like this::

        @wrap_fs_methods(mydecorator)
        class MyFSClass(FS):
            ...

    """
    def apply_decorator(cls):
        for method_name in wrap_fs_methods.method_names:
            if method_name in exclude:
                continue
            method = getattr(cls,method_name,None)
            if method is not None:
                setattr(cls,method_name,decorator(method))
        return cls
    if cls is not None:
        return apply_decorator(cls)
    else:
        return apply_decorator

wrap_fs_methods.method_names = ["open","exists","isdir","isfile","listdir",
    "makedir","remove","setcontents","removedir","rename","getinfo","copy",
    "move","copydir","movedir","close","getxattr","setxattr","delxattr",
    "listxattrs","validatepath","getsyspath","createfile", "hasmeta", "getmeta","listdirinfo",
    "ilistdir","ilistdirinfo"]





#-------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   09/15/2005
#
#-------------------------------------------------------------------------

""" A Traits UI demo that borrows heavily from the design of the wxPython demo.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

import sys
import glob
import token
import tokenize
import operator
from StringIO import StringIO
from configobj import ConfigObj

from traits.api import (HasTraits, HasPrivateTraits, Str, Instance, Property,
                        Any, Code, HTML, true, false, Dict)

from traitsui.api import (TreeEditor, ObjectTreeNode, TreeNodeObject, View,
                          Item, VSplit, Tabbed, VGroup, HGroup, Heading,
                          Handler, UIInfo, InstanceEditor, HTMLEditor, Include,
                          spring)

from os import listdir

from os.path import (join, isdir, split, splitext, dirname, basename, abspath,
                     exists, isabs)

#-------------------------------------------------------------------------
#  Global data:
#-------------------------------------------------------------------------

# Define the code used to populate the 'execfile' dictionary:
exec_str =  """from traits.api import *

"""

#----------------------------------------------------------------------------
#  Return a 'user-friendly' name for a specified string:
#----------------------------------------------------------------------------


def user_name_for(name):
    name = name.replace('_', ' ')
    return name[:1].upper() + name[1:]

#-------------------------------------------------------------------------
#  Parses the contents of a specified source file into module comment and
#  source text:
#-------------------------------------------------------------------------


def extract_docstring_from_source(source):
    """Return module docstring and source code from python source code.

    Returns
    -------
    docstring : str
        The first module-level string; i.e. the module docstring.
    source : str
        The source code, sans docstring.
    """
    # Reset file and generate python tokens
    f = StringIO(source)
    python_tokens = tokenize.generate_tokens(f.readline)

    for ttype, tstring, tstart, tend, tline in python_tokens:
        token_name = token.tok_name[ttype]
        if token_name == 'STRING' and tstart[1] == 0:
            break
    else:
        # No docstrings found. Return blank docstring and all the source.
        return '', source.strip()

    source_lines = source.splitlines()

    # Extract module docstring lines and recombine
    docstring = eval('\n'.join(source_lines[tstart[0] - 1:tend[0]]))
    source_lines = source_lines[:tstart[0] - 1] + source_lines[tend[0]:]
    source = '\n'.join(source_lines)
    source = source.strip()

    return docstring, source


def parse_source(file_name):
    """Return module docstring and source code from python source file.

    Returns
    -------
    docstring : str
        The first module-level string; i.e. the module docstring.
    source : str
        The source code, sans docstring.
    """
    try:
        fh = open(file_name, 'rb')
        source_code = fh.read()
        return extract_docstring_from_source(source_code)
    except:
        return ('', '')


#-------------------------------------------------------------------------
#  'DemoFileHandler' class:
#-------------------------------------------------------------------------

class DemoFileHandler(Handler):

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # The current 'info' object (for use by the 'write' method):
    info = Instance(UIInfo)

    #-------------------------------------------------------------------------
    #  Initializes the view:
    #-------------------------------------------------------------------------

    def init(self, info):
        # Save the reference to the current 'info' object:
        self.info = info

        # Set up the 'print' logger:
        df = info.object
        df.log = ''
        sys.stdout = sys.stderr = self

        # Read in the demo source file:
        df.description, df.source = parse_source(df.path)
        # Try to run the demo source file:

        # Append the path for the demo source file to sys.path, so as to
        # resolve any local (relative) imports in the demo source file.
        sys.path.append(dirname(df.path))

        locals = df.parent.init_dic
        locals['__name__'] = '___main___'
        locals['__file__'] = df.path
        sys.modules['__main__'].__file__ = df.path
        try:
            execfile(df.path, locals, locals)
            demo = self._get_object('modal_popup', locals)
            if demo is not None:
                demo = ModalDemoButton(demo=demo)
            else:
                demo = self._get_object('popup', locals)
                if demo is not None:
                    demo = DemoButton(demo=demo)
                else:
                    demo = self._get_object('demo', locals)
            # FIXME: If a 'demo' object could not be found, then try to execute
            # the file setting __name__ to __main__. A lot of test scripts have
            # the actual test running when __name__==__main__ and so we can at
            # least run all test examples this way. Use a do_later loop so as to
            # finish building the current UI before running the test.
            if demo is None:
                locals['__name__'] = '__main__'
                #do_later(self.execute_test, df, locals)
        except Exception as excp:
            demo = DemoError(msg=str(excp))

        # Clean up sys.path
        sys.path.remove(dirname(df.path))
        df.demo = demo

    def execute_test(self, df, locals):
        """ Executes the file in df.path in the namespace of locals."""
        execfile(df.path, locals, locals)

    #-------------------------------------------------------------------------
    #  Closes the view:
    #-------------------------------------------------------------------------

    def closed(self, info, is_ok):
        """ Closes the view.
        """
        info.object.demo = None

    #-------------------------------------------------------------------------
    #  Get a specified object from the execution dictionary:
    #-------------------------------------------------------------------------

    def _get_object(self, name, dic):
        object = dic.get(name) or dic.get(name.capitalize())
        if object is not None:
            if isinstance(type(object), type):
                try:
                    object = object()
                except:
                    pass

            if isinstance(object, HasTraits):
                return object

        return None

    #-------------------------------------------------------------------------
    #  Handles 'print' output:
    #-------------------------------------------------------------------------

    def write(self, text):
        self.info.object.log += text

    def flush(self):
        pass

# Create a singleton instance:
demo_file_handler = DemoFileHandler()

#-------------------------------------------------------------------------
#  'DemoError' class:
#-------------------------------------------------------------------------


class DemoError(HasPrivateTraits):

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # The error message text:
    msg = Code

    #-------------------------------------------------------------------------
    #  Traits view definitions:
    #-------------------------------------------------------------------------

    traits_view = View(
        VGroup(
            Heading('Error in source file'),
            Item('msg', style='custom', show_label=False),
        )
    )

#-------------------------------------------------------------------------
#  'DemoButton' class:
#-------------------------------------------------------------------------


class DemoButton(HasPrivateTraits):

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # The demo to be launched via a button:
    demo = Instance(HasTraits)

    # The demo view item to use:
    demo_item = Item('demo',
                     show_label=False,
                     editor=InstanceEditor(label='Run demo...', kind='live')
                     )

    #-------------------------------------------------------------------------
    #  Traits view definitions:
    #-------------------------------------------------------------------------

    traits_view = View(
        VGroup(
            VGroup(
                Heading('Click the button to run the demo:'),
                '20'
            ),
            HGroup(
                spring,
                Include('demo_item'),
                spring
            )
        ),
        resizable=True
    )

#-------------------------------------------------------------------------
#  'ModalDemoButton' class:
#-------------------------------------------------------------------------


class ModalDemoButton(DemoButton):

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # The demo view item to use:
    demo_item = Item('demo',
                     show_label=False,
                     editor=InstanceEditor(label='Run demo...', kind='modal')
                     )

#-------------------------------------------------------------------------
#  'DemoTreeNodeObject' class:
#-------------------------------------------------------------------------


class DemoTreeNodeObject(TreeNodeObject):

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Cached result of 'tno_has_children':
    _has_children = Any

    # Cached result of 'tno_get_children':
    _get_children = Any

    #-------------------------------------------------------------------------
    #  Returns whether chidren of this object are allowed or not:
    #-------------------------------------------------------------------------

    def tno_allows_children(self, node):
        """ Returns whether chidren of this object are allowed or not.
        """
        return self.allows_children

    #-------------------------------------------------------------------------
    #  Returns whether or not the object has children:
    #-------------------------------------------------------------------------

    def tno_has_children(self, node=None):
        """ Returns whether or not the object has children.
        """
        if self._has_children is None:
            self._has_children = self.has_children()

        return self._has_children

    #-------------------------------------------------------------------------
    #  Gets the object's children:
    #-------------------------------------------------------------------------

    def tno_get_children(self, node):
        """ Gets the object's children.
        """
        if self._get_children is None:
            self._get_children = self.get_children()

        return self._get_children

    #-------------------------------------------------------------------------
    #  Returns whether or not the object has children:
    #-------------------------------------------------------------------------

    def has_children(self, node):
        """ Returns whether or not the object has children.
        """
        raise NotImplementedError

    #-------------------------------------------------------------------------
    #  Gets the object's children:
    #-------------------------------------------------------------------------

    def get_children(self, node):
        """ Gets the object's children.
        """
        raise NotImplementedError

#-------------------------------------------------------------------------
#  'DemoFile' class:
#-------------------------------------------------------------------------


class DemoFile(DemoTreeNodeObject):

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Parent of this file:
    parent = Any

    # Name of file system path to this file:
    path = Property

    # Name of the file:
    name = Str

    # UI form of the 'name':
    nice_name = Property

    # Files don't allow children:
    allows_children = false

    # Description of what the demo does:
    description = HTML

    # Source code for the demo:
    source = Code

    # Demo object whose traits UI is to be displayed:
    demo = Instance(HasTraits)

    # Log of all print messages displayed:
    log = Code

    _nice_name = Str
    #-------------------------------------------------------------------------
    #  Implementation of the 'path' property:
    #-------------------------------------------------------------------------

    def _get_path(self):
        return join(self.parent.path, self.name + '.py')

    #-------------------------------------------------------------------------
    #  Implementation of the 'nice_name' property:
    #-------------------------------------------------------------------------

    def _get_nice_name(self):
        if not self._nice_name:
            self._nice_name = user_name_for(self.name)
        return self._nice_name

    def _set_nice_name(self, value):
        old = self.nice_name
        self._nice_name = value
        self.trait_property_changed('nice_name', old, value)

    #-------------------------------------------------------------------------
    #  Returns whether or not the object has children:
    #-------------------------------------------------------------------------

    def has_children(self):
        """ Returns whether or not the object has children.
        """
        return False

#-------------------------------------------------------------------------
#  'DemoPath' class:
#-------------------------------------------------------------------------


class DemoPath(DemoTreeNodeObject):

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Parent of this package:
    parent = Any

    # Name of file system path to this package:
    path = Property

    # Name of the directory:
    name = Str

    # UI form of the 'name':
    nice_name = Property

    # Description of the contents of the directory:
    description = Property(HTML)

    # Source code contained in the '__init__.py' file:
    source = Property(Code)

    # Dictionary containing symbols defined by the path's '__init__.py' file:
    init_dic = Property

    # Should .py files be included?
    use_files = true

    # Paths do allow children:
    allows_children = true

    # Configuration dictionary for this node
    # This trait is set when a config file exists for the parent of this path.
    config_dict = Dict

    # Configuration file for this node.
    config_filename = Str

    # Cached value of the nice_name property.
    _nice_name = Str

    #-------------------------------------------------------------------------
    #  Implementation of the 'path' property:
    #-------------------------------------------------------------------------

    def _get_path(self):
        return join(self.parent.path, self.name)

    #-------------------------------------------------------------------------
    #  Implementation of the 'nice_name' property:
    #-------------------------------------------------------------------------

    def _get_nice_name(self):
        if not self._nice_name:
            self._nice_name = user_name_for(self.name)
        return self._nice_name

    #-------------------------------------------------------------------------
    #  Setter for the 'nice_name' property:
    #-------------------------------------------------------------------------

    def _set_nice_name(self, value):
        old = self.nice_name
        self._nice_name = value
        self.trait_property_changed('nice_name', old, value)

    #-------------------------------------------------------------------------
    #  Implementation of the 'description' property:
    #-------------------------------------------------------------------------

    def _get_description(self):
        if self._description is None:
            self._get_init()

        return self._description

    #-------------------------------------------------------------------------
    #  Implementation of the 'source' property:
    #-------------------------------------------------------------------------

    def _get_source(self):
        if self._source is None:
            self._get_init()

        return self._source

    #-------------------------------------------------------------------------
    #  Implementation of the 'init_dic' property:
    #-------------------------------------------------------------------------

    def _get_init_dic(self):
        init_dic = {}
        description, source = parse_source(join(self.path, '__init__.py'))
        exec (exec_str + source) in init_dic
        return init_dic

        # fixme: The following code should work, but doesn't, so we use the
        #        preceding code instead. Changing any trait in the object in
        #        this method causes the tree to behave as if the DemoPath object
        #        had been selected instead of a DemoFile object. May be due to
        #        an 'anytrait' listener in the TreeEditor?
        #if self._init_dic is None:
        #   self._init_dic = {}
        #   #exec self.source in self._init_dic
        #return self._init_dic.copy()

    #-------------------------------------------------------------------------
    #  Initializes the description and source from the path's '__init__.py'
    #  file:
    #-------------------------------------------------------------------------

    def _get_init(self):
        if self.use_files:
            # Read in the '__init__.py' source file (if any):
            self._description, source = parse_source(
                join(self.path, '__init__.py'))
        else:
            self._description = ('<img src="traits_ui_demo.jpg">')
            source = ''

        self._source = exec_str + source

    #-------------------------------------------------------------------------
    #  Returns whether or not the object has children:
    #-------------------------------------------------------------------------

    def has_children(self):
        """ Returns whether or not the object has children.
        """
        path = self.path
        for name in listdir(path):
            cur_path = join(path, name)
            if isdir(cur_path):
                return True

            if self.use_files:
                name, ext = splitext(name)
                if (ext == '.py') and (name != '__init__'):
                    return True

        return False

    #-------------------------------------------------------------------------
    #  Gets the object's children:
    #-------------------------------------------------------------------------

    def get_children(self):
        """ Gets the object's children.
        """
        if self.config_dict or self.config_filename:
            children = self.get_children_from_config()
        else:
            children = self.get_children_from_datastructure()
        return children

    #-------------------------------------------------------------------------
    #  Gets the object's children based on the filesystem structure.
    #-------------------------------------------------------------------------
    def get_children_from_datastructure(self):
        """ Gets the object's children based on the filesystem structure.
        """

        dirs = []
        files = []
        path = self.path
        for name in listdir(path):
            cur_path = join(path, name)
            if isdir(cur_path):
                if self.has_py_files(cur_path):
                    dirs.append(DemoPath(parent=self, name=name))

            elif self.use_files:
                name, ext = splitext(name)
                if (ext == '.py') and (name != '__init__'):
                    files.append(DemoFile(parent=self, name=name))

        sort_key = operator.attrgetter('name')
        dirs.sort(key=sort_key)
        files.sort(key=sort_key)

        return (dirs + files)

    #-------------------------------------------------------------------------
    # Gets the object's children as specified in its configuration file or
    # dictionary.
    #-------------------------------------------------------------------------

    def get_children_from_config(self):
        """
        Gets the object's children as specified in its configuration file or
        dictionary.
        """

        if not self.config_dict:
            if exists(self.config_filename):
                try:
                    self.config_dict = ConfigObj(self.config_filename)
                except:
                    pass
        if not self.config_dict:
            return self.get_children_from_datastructure()

        dirs = []
        files = []
        for keyword, value in self.config_dict.items():
            if not value.get('no_demo'):
                sourcedir = value.pop('sourcedir', None)
                if sourcedir is not None:
                    # This is a demo directory.
                    demoobj = DemoPath(parent=self, name=sourcedir)
                    demoobj.nice_name = keyword
                    demoobj.config_dict = value
                    dirs.append(demoobj)
                else:
                    names = []
                    filenames = value.pop('files', [])
                    if not isinstance(filenames, list):
                        filenames = [filenames]
                    for filename in filenames:
                        filename = join(self.path, filename)
                        for name in glob.iglob(filename):
                            pathname, ext = splitext(name)
                            if (ext == '.py') and \
                                    (basename(pathname) != '__init__'):
                                names.append(pathname)
                    if len(names) > 1:
                        config_dict = {}
                        for name in names:
                            config_dict[basename(name)] = {'files':
                                                           name + '.py'}
                        demoobj = DemoPath(parent=self, name='')
                        demoobj.nice_name = keyword
                        demoobj.config_dict = config_dict
                        dirs.append(demoobj)
                    elif len(names) == 1:
                        file = DemoFile(parent=self, name=names[0])
                        file.nice_name = keyword
                        files.append(file)

        sort_key = operator.attrgetter('nice_name')
        dirs.sort(key=sort_key)
        files.sort(key=sort_key)

        return (dirs + files)

    #-------------------------------------------------------------------------
    #  Returns whether the specified path contains any .py files:
    #-------------------------------------------------------------------------

    def has_py_files(self, path):
        for name in listdir(path):
            cur_path = join(path, name)
            if isdir(cur_path):
                if self.has_py_files(cur_path):
                    return True

            else:
                name, ext = splitext(name)
                if ext == '.py':
                    return True

        return False

#-------------------------------------------------------------------------
#  Defines the demo tree editor:
#-------------------------------------------------------------------------

path_view = View(
    Tabbed(
        Item('description',
             label='Description',
             show_label=False,
             style='readonly',
             editor=HTMLEditor(format_text=True)
             ),
        Item('source',
             label='Source',
             show_label=False,
             style='custom'
             ),
        export='DockWindowShell',
        id='tabbed'
    ),
    id='traitsui.demos.demo.path_view',
    #dock    = 'horizontal'
)

demo_view = View(
    #VSplit(
    Tabbed(
        Item('description',
             label='Description',
             show_label=False,
             style='readonly',
             editor=HTMLEditor(format_text=True)
             ),
        Item('source',
             label='Source',
             show_label=False,
             style='custom'
             ),
        Item('demo',
             label='Demo',
             show_label=False,
             style='custom',
             resizable=True,
             # FIXME:
             # visible_when doesn't work correctly yet (for wx atleast)
             # for tabbed items. Needs more investigation.
                   visible_when='demo',
             ),
        Item('log',
             show_label=False,
             style='readonly'
             ),
        export='DockWindowShell',
        id='tabbed',
    ),
    # JDM moving log panel provisionally to its own tab, distracting here.
    #VGroup(
    #Item( 'log',
    #show_label = False,
    #style      = 'readonly'
    #),
    #label = 'Log'
    #),
    #export = 'DockWindowShell',
    #id     = 'vsplit'
    #),
    id='traitsui.demos.demo.file_view',
    #dock    = 'horizontal',
    handler=demo_file_handler
)

demo_tree_editor = TreeEditor(
    nodes=[
        ObjectTreeNode(node_for=[DemoPath],
                       label='nice_name',
                       view=path_view),
        ObjectTreeNode(node_for=[DemoFile],
                       label='nice_name',
                       view=demo_view)
    ]
)

#-------------------------------------------------------------------------
#  'Demo' class:
#-------------------------------------------------------------------------


class Demo(HasPrivateTraits):

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Path to the root demo directory:
    path = Str

    # Root path object for locating demo files:
    root = Instance(DemoPath)

    # Title for the demo
    title = Str

    #-------------------------------------------------------------------------
    #  Traits view definitions:
    #-------------------------------------------------------------------------

    def default_traits_view(self):
        """ Constructs the default traits view."""

        traits_view = View(
            Item(name='root',
                 id='root',
                 show_label=False,
                 editor=demo_tree_editor
                 ),
            title=self.title,
            id='traitsui.demos.demo.Demo',
            #dock      = 'horizontal',
            resizable=True,
            # JDM: Seems that QT interface does not deal well with these size
            # limits.
            # With them, we get repeated:
            #   Object::disconnect: Parentheses expected, signal AdvancedCodeWidget::lostFocus
            # But without them, it throws an exception on exit:
            #    Internal C++ object (_StickyDialog) already deleted.
            # No, actually sometimes we get the latter even with them.
            width=950,
            height=900
        )
        return traits_view

    #-------------------------------------------------------------------------
    #  Handles the 'root' trait being changed:
    #-------------------------------------------------------------------------

    def _root_changed(self, root):
        """ Handles the 'root' trait being changed.
        """
        root.parent = self

#-------------------------------------------------------------------------
#  Function to run the demo:
#-------------------------------------------------------------------------


def demo(use_files=False, dir_name=None, config_filename='',
         title='Traits UI Demos'):
    if dir_name is None:
        dir_name = dirname(abspath(sys.argv[0]))
    path, name = split(dir_name)
    if len(config_filename) > 0 and not isabs(config_filename):
        config_filename = join(path, name, config_filename)
    Demo(path=path,
         title=title,
         root=DemoPath(name=name,
                       use_files=use_files,
                       config_filename=config_filename)
         ).configure_traits()



#-------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   12/14/2005
#
#-------------------------------------------------------------------------

""" Defines the DockableViewElement class, which allows Traits UIs and
    Traits UI elements to be docked in external PyFace DockWindow windows.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import HasPrivateTraits, Instance, Bool

from .ui import UI

from .group import Group

from .view import View

from .view_element import ViewSubElement

from pyface.dock.idockable import IDockable

#-------------------------------------------------------------------------
#  'DockableViewElement' class:
#-------------------------------------------------------------------------


class DockableViewElement(HasPrivateTraits, IDockable):
    """ Allows Traits UIs and Traits UI elements to be docked in external
        PyFace DockWindow windows.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # The Traits UI that can be docked with an external DockWindow
    ui = Instance(UI)

    # The (optional) element of the Traits UI that can be docked
    element = Instance(ViewSubElement)

    # Should the DockControl be closed on redocking?
    should_close = Bool(False)

#-- IDockable interface --------------------------------------------------

    #-------------------------------------------------------------------------
    #  Should the current DockControl be closed before creating the new one:
    #-------------------------------------------------------------------------

    def dockable_should_close(self):
        """ Should the current DockControl be closed before creating the new
            one?
        """
        element = self.element
        if element is None:
            element = self.ui.view.content

        if not isinstance(element, Group):
            element = Group().set(content=[element])

        group = Group().set(content=[element])
        self._view = View().set(**self.ui.view.get()).set(content=group,
                                                          title='')

        # FIXME: The following private traits are being set here to facilitate
        # rebuilding the ui (which will require the context and the handler).
        # When a current dock control is closed (close_dock_control method), the
        # contents of self.ui have been disposed of and self.ui is now None.
        # Now if a new UI needs to be created by calling dockable_get_control
        # (e.g., when doing an 'undock' action on a dock window), we need to
        # pass on the context and handler to the UI. Therefore, we are setting
        # these private traits here so dockable_get_control can access them.
        # In future, we need to investigate if there is a better way to do
        # this.

        self._context = self.ui.context.copy()
        # Make copy since context will be emptied when calling
        # self.ui.dispose()
        self._handler = self.ui.handler

        return (self.should_close or (self.element is None))

    #-------------------------------------------------------------------------
    #  Gets a control that can be docked into a DockWindow:
    #-------------------------------------------------------------------------

    def dockable_get_control(self, parent):
        """ Gets a control that can be docked into a DockWindow.
        """
        # Create the new UI:
        ui = self._view.ui(self._context, parent=parent,
                           kind='subpanel',
                           handler=self._handler)

        # Discard the reference to the view created previously:
        self._view = None

        # If the old UI was closed, then switch to using the new one:
        if self.element is None:
            self.ui = ui
        else:
            self._ui = ui

        return ui.control

    #-------------------------------------------------------------------------
    #  Allows the object to override the default DockControl settings:
    #-------------------------------------------------------------------------

    def dockable_init_dockcontrol(self, dock_control):
        """ Allows the object to override the default DockControl settings.
        """
        dockable = self
        if self.element is not None:
            dockable = DockableViewElement(ui=self._ui,
                                           element=self.element,
                                           should_close=True)
            self._ui = None

        dock_control.set(dockable=dockable,
                         on_close=dockable.close_dock_control)

    #-------------------------------------------------------------------------
    #  Handles the closing of a DockControl containing a Traits UI:
    #-------------------------------------------------------------------------

    def close_dock_control(self, dock_control, abort):
        """ Handles the closing of a DockControl containing a Traits UI.
        """
        ui = self.ui

        # Ask the traits UI handler if it is OK to close the window:
        if (not abort) and (not ui.handler.close(ui.info, True)):
            # If not, tell the DockWindow not to close it:
            return False

        # Otherwise, clean up and close the traits UI:
        ui.dispose(abort=abort)

        # Break our linkage to the UI and ViewElement object:
        self.ui = self.element = None

        # And tell the DockWindow to remove the DockControl:
        return True



#------------------------------------------------------------------------------
# Copyright (c) 2008, Enthought, Inc.
# All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#
#------------------------------------------------------------------------------
""" Defines the editor factory for single-selection enumerations, for all traits
    user interface toolkits.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

import os
import sys

from ..editor_factory import EditorWithListFactory

from traits.api import Any, Range, Enum, Bool

from ..toolkit import toolkit_object

#-------------------------------------------------------------------------
#  Trait definitions:
#-------------------------------------------------------------------------

# Supported display modes for a custom style editor
Mode = Enum('radio', 'list')

#-------------------------------------------------------------------------
#  'ToolkitEditorFactory' class:
#-------------------------------------------------------------------------


class ToolkitEditorFactory(EditorWithListFactory):
    """ Editor factory for enumeration editors.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # (Optional) Function used to evaluate text input:
    evaluate = Any

    # Is user input set on every keystroke (when text input is allowed)?
    auto_set = Bool(True)

    # Number of columns to use when displayed as a grid:
    cols = Range(1, 20)

    # Display modes supported for a custom style editor:
    mode = Mode

    #-------------------------------------------------------------------------
    #  'Editor' factory methods:
    #-------------------------------------------------------------------------

    def _get_custom_editor_class(self):
        """ Returns the editor class to use for "custom" style views.
        Overridden to return the editor class for the specified mode.
        """
        editor_file_name = \
            os.path.basename(sys.modules[self.__class__.__module__].
                             __file__)
        try:
            if self.mode == 'radio':
                return toolkit_object(editor_file_name.split('.')[0] +
                                      ':RadioEditor',
                                      raise_exceptions=True)
            else:
                return toolkit_object(editor_file_name.split('.')[0] +
                                      ':ListEditor',
                                      raise_exceptions=True)
        except:
            return super(ToolkitEditorFactory, self)._get_custom_editor_class()


# Define the EnumEditor class.
EnumEditor = ToolkitEditorFactory

### EOF #######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought application scripting package component>
#------------------------------------------------------------------------------


# Enthought library imports.
from traits.api import Bool, Event, Instance, Interface, Unicode

# Local imports.
from i_bind_event import IBindEvent


class IScriptManager(Interface):
    """ The script manager interface.  A script manager is responsible for the
    recording of appropriately annotated user actions as scripts that can be
    executed without user intervention at a later time.  Typically an
    application would have a single script manager.
    """

    #### 'IScriptManager' interface ###########################################

    # This event is fired whenever a scriptable object is bound or unbound.  It
    # is intended to be used by an interactive Python shell to give the
    # advanced user access to the scriptable objects.  If an object is created
    # via a factory then the event is fired when the factory is called, and not
    # when the factory is bound.
    bind_event = Event(IBindEvent)

    # This is set if user actions are being recorded as a script.  It is
    # maintained by the script manager.
    recording = Bool(False)

    # This is the text of the script currently being recorded (or the last
    # recorded script if none is currently being recorded).  It is updated
    # automatically as the user performs actions.
    script = Unicode

    # This event is fired when the recorded script changes.  The value of the
    # event will be the ScriptManager instance.
    script_updated = Event(Instance('apptools.appscripting.api.IScriptManager'))

    ###########################################################################
    # 'IScriptManager' interface.
    ###########################################################################

    def bind(self, obj, name=None, bind_policy='unique', api=None,
            includes=None, excludes=None):
        """ Bind obj to name and make (by default) its public methods and
        traits (ie. those not beginning with an underscore) scriptable.  The
        default value of name is the type of obj with the first character
        forced to lower case.

        bind_policy determines what happens if the name is already bound.  If
        the policy is 'auto' then a numerical suffix will be added to the name,
        if necessary, to make it unique.  If the policy is 'unique' then an
        exception is raised.  If the policy is 'rebind' then the previous
        binding is discarded.  The default is 'unique'

        If api is given then it is a class, or a list of classes, that define
        the attributes that will be made scriptable.

        Otherwise if includes is given it is a list of names of attributes that
        will be made scriptable.

        Otherwise all the public attributes of scripted_type will be made
        scriptable except those in the excludes list.
        """

    def bind_factory(self, factory, name, bind_policy='unique', api=None,
            includes=None, excludes=None):
        """ Bind factory to name.  This does the same as the bind() method
        except that it uses a factory that will be called later on to create
        the object only if the object is needed.

        See the documentation for bind() for a description of the remaining
        arguments.
        """

    def run(self, script):
        """ Run the given script, either a string or a file-like object.
        """

    def run_file(self, file_name):
        """ Run the given script file.
        """

    def start_recording(self):
        """ Start the recording of user actions.  The 'script' trait is cleared
        and all subsequent actions are added to 'script'.  The 'recording'
        trait is updated appropriately.
        """

    def stop_recording(self):
        """ Stop the recording of user actions.  The 'recording' trait is
        updated appropriately.
        """



import math
import re
import string

import graphics
import lines
import traceback

import OpenGL.GL as gl

from svg_parser_utils import parse_float, parse_list, get_fns
from svg_path_builder import SVGPathBuilder

from glutils import DisplayListGenerator
import svg_style
from vector_math import Matrix, BoundingBox, vec2
from svg_constants import XMLNS
from render_target import CanvasManager


class SVGContainer(object):

    def __init__(self, parent):

        self.parent = parent

        self.children = []

        self.is_def = False

        if parent:
            parent.add_child(self)

    def add_child(self, child):
        self.children.append(child)


class SVGRenderableElement(SVGContainer):

    def __init__(self, svg, element, parent):

        #: The id of the element
        self.id = element.get('id', '')

        SVGContainer.__init__(self, parent)

        self.svg = svg

        #: Is this element a pattern?
        self.is_pattern = element.tag.endswith('pattern')

        #: Is this element a definition?
        self.is_def = False

        if parent:
            self.is_def = parent.is_def

        #: The element style (possibly with inherited traits from parent style)
        self.style = svg_style.SVGStyle(
            parent.style if isinstance(parent, SVGRenderableElement) else None)
        self.style.from_element(element)

        #: Optional element title
        self.title = element.findtext('{%s}title' % (XMLNS,))

        #: Optional element description. Useful for embedding metadata.
        self.description = element.findtext('{%s}desc' % (XMLNS,))

        #construct a matrix for each transform
        t_acc = Matrix.identity()

        transform_str = element.get('transform', None)
        if transform_str:
            transforms = get_fns(transform_str)

            for tstring in transforms:
                t_acc = t_acc * Matrix(tstring)

        #: Element transforms
        self.transform = t_acc

        #: Children elements
        self.children = []

        #: XML tag this was originally.
        self.tag_type = element.tag

    def add_child(self, child):
        """Add a child to this element class (usually children register with parent)"""
        self.children.append(child)

    @property
    def is_pattern_part(self):
        part = self
        while part:
            if isinstance(part, SVGRenderableElement) and part.is_pattern:
                return True
            part = part.parent
        return False

    @property
    def absolute_transform(self):
        """Return this transform, multiplied by chain of parents"""
        if self.parent:
            return self.parent.absolute_transform * self.transform
        return self.transform

    def on_render(self):
        pass

    def render(self):
        #with CanvasManager.inst().temp():
            with self.transform:
                self.on_render()

                for c in self.children:
                    c.render()


class SVGGroup(SVGRenderableElement):
    pass


class SVGMarker(SVGRenderableElement):

    def __init__(self, svg, element, parent):
        SVGRenderableElement.__init__(self, svg, element, parent)

        self.units = element.get('markerUnits', 'strokeWidth')
        self.marker_width = parse_float(element.get('markerWidth', '3'))
        self.marker_height = parse_float(element.get('markerHeight', '3'))
        self.orient = element.get('orient', 'auto')
        self.ref_x = parse_float(element.get('refX', '0'))
        self.ref_y = parse_float(element.get('refY', '0'))

        vb = element.get('viewBox', None)

        if vb:
            x, y, w, h = (parse_float(x) for x in parse_list(element.get("viewBox")))
            self.vb_x, self.vb_y, self.vb_w, self.vb_h = x, y, w, h
        else:
            self.vb_x = 0
            self.vb_y = 0
            self.vb_w = 1
            self.vb_h = 1


XLINK_NS = "{http://www.w3.org/1999/xlink}"


class SVGUse(SVGRenderableElement):
    """Represents an SVG "use" directive, to reuse a predefined path"""

    def __init__(self, svg, element, parent):
        SVGRenderableElement.__init__(self, svg, element, parent)
        self.svg = svg
        self.target = element.get(XLINK_NS + "href", None)
        self.x = parse_float(element.get('x', '0'))
        self.y = parse_float(element.get('y', '0'))

        self.transform = self.transform * Matrix.translation(self.x, self.y)

        #clip off "#"
        if self.target:
            self.target = self.target[1:]

    def render(self):
        with self.transform:
            defn = self.svg.defs[self.target]
            defn.render()


class SVGDefs(SVGRenderableElement):
    """Represents an SVG "defs" directive, to define paths without drawing them"""

    def __init__(self, svg, element, parent):
        SVGRenderableElement.__init__(self, svg, element, parent)
        self.svg = svg
        self.is_def = True

    def add_child(self, child):
        if hasattr(child, 'id'):
            self.svg.defs[child.id] = child
        self.children.append(child)


def flatten_list(l):
    new_list = []
    for x in l:
        new_list.extend(x)
    return new_list


class SVGPath(SVGRenderableElement):
    """
    Represents a single SVG path. This is usually
    a distinct shape with a fill pattern,
    an outline, or both.
    """

    def __init__(self, svg, element, parent):

        SVGRenderableElement.__init__(self, svg, element, parent)

        #: The original SVG file
        self.svg = svg

        if not self.is_pattern_part:
            self.config = svg.config
        else:
            self.config = svg.config.super_detailed()

        #: The actual path elements, as a list of vertices
        self.outlines = None

        #: The triangles that comprise the inner fill
        self.triangles = None

        #: The base shape. Possible values: path, rect, circle, ellipse, line, polygon, polyline
        self.shape = None

        #: The bounding box
        self._bounding_box = None

        self.marker_start = element.get('marker-start', None)
        self.marker_mid = element.get('marker-mid', None)
        self.marker_end = element.get('marker-end', None)

        if self.marker_start: self.marker_start = self.marker_start[5:-1]
        if self.marker_mid: self.marker_mid = self.marker_mid[5:-1]
        if self.marker_end: self.marker_end = self.marker_end[5:-1]

        path_builder = SVGPathBuilder()

        path_builder.read_xml_svg_element(
                        self,
                        element,
                        self.config)

        self.outlines = path_builder.path

        self.triangles = path_builder.polygon

        self.display_list = None

    def _render_stroke(self):
        stroke = self.style.stroke
        stroke_width = self.style.stroke_width

        is_miter = self.style.stroke_linejoin == 'miter'

        miter_limit = self.style.stroke_miterlimit if is_miter else 0

        for loop in self.outlines:
            self.svg.n_lines += len(loop) - 1
            loop_plus = []

            for i in xrange(len(loop) - 1):
                loop_plus += [loop[i], loop[i+1]]

            if isinstance(stroke, str):
                g = self.svg._gradients[stroke]
                strokes = [g.sample(x, self) for x in loop_plus]
            else:
                strokes = [stroke for x in loop_plus]

            if len(loop_plus) == 0:
                continue

            if len(self.style.stroke_dasharray):
                ls = lines.split_line_by_pattern(loop_plus, self.style.stroke_dasharray)

                if ls[0][0] == ls[-1][-1]:
                    #if the last line end point equals the first line start point,
                    #this is a "closed" line, so combine the first and the last line
                    combined_line = ls[-1] + ls[0]
                    ls[0] = combined_line
                    del ls[-1]

                for l in ls:
                    lines.draw_polyline(
                        l,
                        stroke_width,
                        color=strokes[0],
                        line_cap=self.style.stroke_linecap,
                        join_type=self.style.stroke_linejoin,
                        miter_limit=miter_limit)

                if self.marker_start:
                    end_point = vec2(loop_plus[0])
                    almost_end_point = vec2(loop_plus[1])
                    marker = self.svg.defs[self.marker_start]
                    self._render_marker(end_point, almost_end_point, marker, True)
                if self.marker_end:
                    end_point = vec2(loop_plus[-1])
                    almost_end_point = vec2(loop_plus[-2])
                    marker = self.svg.defs[self.marker_end]
                    self._render_marker(end_point, almost_end_point, marker)

            else:
                lines.draw_polyline(
                    loop_plus,
                    stroke_width,
                    color=strokes[0],
                    line_cap=self.style.stroke_linecap,
                    join_type=self.style.stroke_linejoin,
                    miter_limit=miter_limit)

                if self.marker_start:
                    end_point = vec2(loop_plus[0])
                    almost_end_point = vec2(loop_plus[1])
                    marker = self.svg.defs[self.marker_start]
                    self._render_marker(end_point, almost_end_point, marker, True)
                if self.marker_end:
                    end_point = vec2(loop_plus[-1])
                    almost_end_point = vec2(loop_plus[-2])
                    marker = self.svg.defs[self.marker_end]
                    self._render_marker(end_point, almost_end_point, marker)

    def _render_marker(self, a, b, marker, reverse=False):
        if marker.orient == 'auto':
            angle = (a - b).angle()
        else:
            angle = marker.orient

        if reverse:
            angle += math.pi

        sx = (marker.marker_width / marker.vb_w) * self.style.stroke_width
        sy = (marker.marker_height / marker.vb_h) * self.style.stroke_width

        rx = marker.ref_x
        ry = marker.ref_y

        with Matrix.transform(a.x, a.y, theta=angle):
            with Matrix.scale(sx, sy):
                with Matrix.translation(-rx, -ry):
                    marker.render()

    def _render_gradient_fill(self):
        fill = self.style.fill
        tris = self.triangles
        self.svg.n_tris += len(tris) / 3
        g = None
        if isinstance(fill, str):
            g = self.svg._gradients[fill]
            fills = [g.sample(x, self) for x in tris]
        else:
            fills = [fill] * len(tris)  # for x in tris]

        if g:
            g.apply_shader(self, self.transform, self.style.opacity * self.style.fill_opacity)

        graphics.draw_colored_triangles(
            flatten_list(tris),
            flatten_list(fills)
        )

        if g:
            g.unapply_shader()

    def bounding_box(self):
        '''
        returns a tuple describing the bounding box:

        (min_x, min_y, max_x, max_y)
        '''
        if not self._bounding_box:
            self._bounding_box = BoundingBox()

            if self.triangles:
                self._bounding_box.expand(self.triangles)
            if self.outlines:
                for o in self.outlines:
                    self._bounding_box.expand(o)
        return self._bounding_box.extents()

    def _render_pattern_fill(self):
        fill = self.style.fill
        tris = self.triangles
        pattern = None
        if fill in self.svg.patterns:
            pattern = self.svg.patterns[fill]
            pattern.bind_texture()

        min_x, min_y, max_x, max_y = self.bounding_box()

        tex_coords = []

        for vtx in tris:
            tex_coords.append((vtx[0]-min_x)/(max_x-min_x)/pattern.width)
            tex_coords.append((vtx[1]-min_y)/(max_y-min_y)/pattern.width)

        graphics.draw_textured_triangles(
            flatten_list(tris),
            tex_coords
        )

        if pattern:
            pattern.unbind_texture()

    def on_render(self):
        """Render immediately to screen (no display list). Slow! Consider
        using SVG.draw(...) instead."""

        gl.glClear(gl.GL_DEPTH_BUFFER_BIT)

        gl.glEnable(gl.GL_DEPTH_TEST)

        if self.style.stroke and self.outlines:
            self._render_stroke()

        gl.glPushMatrix()
        gl.glTranslatef(0, 0, -0.1)
        if self.triangles:
            try:
                if isinstance(self.style.fill, str) and self.style.fill in self.svg.patterns:
                    self._render_pattern_fill()
                else:
                    self._render_gradient_fill()
            except Exception as exception:
                traceback.print_exc(exception)
        gl.glPopMatrix()
        gl.glDisable(gl.GL_DEPTH_TEST)



    def __repr__(self):
        return "<SVGPath id=%s title='%s' description='%s' transform=%s>" % (
            self.id, self.title, self.description, self.transform
        )


class SVGViewBox:

    def __init__(self, x, y, w, h):
        pass

    def matrix(self):
        pass


from contextlib import contextmanager
import tempfile
import os

from ..utils import open_h5file


SEPARATOR = '-' * 60


@contextmanager
def temp_file(suffix='', prefix='tmp', dir=None):
    fd, filename = tempfile.mkstemp(suffix=suffix, prefix=prefix, dir=dir)
    try:
        yield filename
    finally:
        os.close(fd)
        os.unlink(filename)


@contextmanager
def temp_h5_file(**kwargs):
    with temp_file(suffix='h5') as fn:
        with open_h5file(fn, mode='a', **kwargs) as h5:
            yield h5



#-----------------------------------------------------------------------------
#
#  Copyright (c) 2006, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: Dave Peterson <dpeterson@enthought.com>
#
#-----------------------------------------------------------------------------

"""
A handler that delegates the handling of events to a set of sub-handlers.

This is typically used as the handler for dynamic views.  See the
**traits.has_dynamic_view** module.
"""

# Enthought library imports
from traits.api import HasTraits, List
from .ui import Dispatcher

# Local imports.
from .handler import Handler

# Set up a logger:
import logging
logger = logging.getLogger(__name__)


class DelegatingHandler(Handler):
    """ A handler that delegates the handling of events to a set of
        sub-handlers.
    """

    #-- Public 'DelegatingHandler' Interface ---------------------------------

    # The list of sub-handlers this object delegates to:
    sub_handlers = List(HasTraits)

    #-- Protected 'DelegatingHandler' Interface ------------------------------

    # A list of dispatchable handler methods:
    _dispatchers = List

    #-------------------------------------------------------------------------
    #  'Handler' interface:
    #-------------------------------------------------------------------------

    #-- Public Methods -------------------------------------------------------

    def closed(self, info, is_ok):
        """ Handles the user interface being closed by the user.

            This method is overridden here to unregister any dispatchers that
            were set up in the *init()* method.
        """
        for d in self._dispatchers:
            d.remove()

    def init(self, info):
        """ Initializes the controls of a user interface.

        Parameters
        ----------
        info : *UIInfo* object
            The UIInfo object associated with the view

        Returns
        -------
        A boolean, indicating whether the user interface was successfully
        initialized. A True value indicates that the UI can be displayed;
        a False value indicates that the display operation should be
        cancelled.

        Description
        -----------
        This method is called after all user interface elements have been
        created, but before the user interface is displayed. Use this method to
        further customize the user interface before it is displayed.

        This method is overridden here to delegate to sub-handlers.
        """

        # Iterate through our sub-handlers, and for each method whose name is
        # of the form 'object_name_changed', where 'object' is the name of an
        # object in the UI's context, create a trait notification handler that
        # will call the method whenever object's 'name' trait changes.
        logger.debug('Initializing delegation in DelegatingHandler [%s]',
                     self)
        context = info.ui.context
        for h in self.sub_handlers:
            # fixme: I don't know why this wasn't here before... I'm not
            # sure this is right!
            h.init(info)

            for name in self._each_trait_method(h):
                if name[-8:] == '_changed':
                    prefix = name[:-8]
                    col = prefix.find('_', 1)
                    if col >= 0:
                        object = context.get(prefix[:col])
                        if object is not None:
                            logger.debug('\tto method [%s] on handler[%s]',
                                         name, h)
                            method = getattr(h, name)
                            trait_name = prefix[col + 1:]
                            self._dispatchers.append(
                                Dispatcher(method, info, object, trait_name)
                            )

                            # Also invoke the method immediately so initial
                            # user interface state can be correctly set.
                            if object.base_trait(trait_name).type != 'event':
                                method(info)

                # fixme: These are explicit workarounds for problems with:-
                #
                # 'GeometryHierarchyViewHandler'
                #
                # which is used in the :-
                #
                # 'GeometryHierarchyTreeEditor'
                #
                # which are in the 'encode.cad.ui.geometry' package.
                #
                # The tree editor has dynamic views, and hence the handler gets
                # wrapped by a 'DelegatingHandler'. Unfortunately the handler
                # has a couple of methods that aren't picked up by the usual
                # wrapping strategy:-
                #
                # 1) 'tree_item_selected'
                #
                # - which is obviously called when a tree item is selected.
                #
                # 2) 'inspect_object'
                #
                # - which is called directly as as action from the context menu
                #   defined in the tree editor.
                #
                elif name in ['tree_item_selected', 'inspect_object']:
                    self.__dict__[name] = self._create_delegate(h, name)

        return True

    def _create_delegate(self, h, name):
        """ Quick fix for handler methods that are currently left out!
        """

        def delegate(*args, **kw):
            method = getattr(h, name)
            return method(*args, **kw)

        return delegate



from __future__ import with_statement

import socket
import threading
from packetstream import JSONDecoder, JSONFileEncoder



class _SocketFile(object):
    def __init__(self, socket):
        self.socket = socket
        
    def read(self, size):
        try:
            return self.socket.recv(size)
        except socket.error:
            return ''
    
    def write(self, data):
        self.socket.sendall(data)


def remote_call(method_name=None):
    method = method_name    
    def deco(f):
        if not hasattr(f, '_remote_call_names'):
            f._remote_call_names = []        
        f._remote_call_names.append(method or f.__name__)
        return f      
    return deco


class RemoteResponse(Exception):
    def __init__(self, header, payload):
        self.header = header
        self.payload = payload

class ConnectionHandlerBase(threading.Thread):
    
    _methods = {}
    
    def __init__(self, server, connection_id, socket, address):
        super(ConnectionHandlerBase, self).__init__()
        self.server = server
        self.connection_id = connection_id        
        self.socket = socket
        self.transport = _SocketFile(socket)
        self.address = address        
        self.encoder = JSONFileEncoder(self.transport)
        self.decoder = JSONDecoder(prelude_callback=self.on_stream_prelude)        
    
        self._lock = threading.RLock()
        self.socket_error = None
          
        if not self._methods:  
            for method_name in dir(self):
                method = getattr(self, method_name)
                if callable(method) and hasattr(method, '_remote_call_names'):
                    for name in method._remote_call_names:
            
                        self._methods[name] = method
                        
        print self._methods
            
        self.fs = None        
    
    def run(self):
        self.transport.write('pyfs/1.0\n')
        while True:
            try:
                data = self.transport.read(4096)
            except socket.error, socket_error:
                print socket_error
                self.socket_error = socket_error
                break
            print "data", repr(data)
            if data: 
                for packet in self.decoder.feed(data):                            
                    print repr(packet)
                    self.on_packet(*packet)
            else:
                break            
        self.on_connection_close()
            
    def close(self):
        with self._lock:
            self.socket.close()
    
    def on_connection_close(self):
        self.socket.shutdown(socket.SHUT_RDWR)
        self.socket.close()
        self.server.on_connection_close(self.connection_id)
    
    def on_stream_prelude(self, packet_stream, prelude):
        print "prelude", prelude
        return True
    
    def on_packet(self, header, payload):
        print '-' * 30
        print repr(header)
        print repr(payload)
        if header['type'] == 'rpc':
            method = header['method']
            args = header['args']
            kwargs = header['kwargs']
            method_callable = self._methods[method]
            remote = dict(type='rpcresult',
                          client_ref = header['client_ref'])
            try:
                response = method_callable(*args, **kwargs)
                remote['response'] = response
                self.encoder.write(remote, '')
            except RemoteResponse, response:
                self.encoder.write(response.header, response.payload)                                

class RemoteFSConnection(ConnectionHandlerBase):

    @remote_call()
    def auth(self, username, password, resource):
        self.username = username
        self.password = password
        self.resource = resource
        from fs.memoryfs import MemoryFS
        self.fs = MemoryFS()
        
class Server(object):
    
    def __init__(self, addr='', port=3000, connection_factory=RemoteFSConnection):
        self.addr = addr
        self.port = port
        self.connection_factory = connection_factory
        self.socket = None
        self.connection_id = 0
        self.threads = {}
        self._lock = threading.RLock()
    
    def serve_forever(self):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind((self.addr, self.port))        
        
        sock.listen(5)
        
        try:
            while True:
                clientsocket, address = sock.accept()
                self.on_connect(clientsocket, address)
        except KeyboardInterrupt:
            pass
        
        try:
            self._close_graceful()
        except KeyboardInterrupt:
            self._close_harsh()
        
    def _close_graceful(self):
        """Tell all threads to exit and wait for them"""
        with self._lock:
            for connection in self.threads.itervalues():
                connection.close()        
            for connection in self.threads.itervalues():
                connection.join()
            self.threads.clear()
            
    def _close_harsh(self):
        with self._lock:
            for connection in self.threads.itervalues():
                connection.close()
            self.threads.clear()
                          
    def on_connect(self, clientsocket, address):
        print "Connection from", address
        with self._lock:
            self.connection_id += 1
            thread = self.connection_factory(self,
                                             self.connection_id,
                                             clientsocket,
                                             address)            
            self.threads[self.connection_id] = thread            
            thread.start()            
        
    def on_connection_close(self, connection_id):
        pass
        #with self._lock:
        #    self.threads[connection_id].join()
        #    del self.threads[connection_id]
            
if __name__ == "__main__":
    server = Server()
    server.serve_forever()
    


#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   12/22/2004
#
#------------------------------------------------------------------------------

""" Trait definition for a wxPython-based color.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------
import wx

from traits.api \
    import Trait, TraitError

# Version dependent imports (ColourPtr not defined in wxPython 2.5):
try:
    ColourPtr = wx.ColourPtr
except:
    class ColourPtr(object):
        pass


#-------------------------------------------------------------------------
#  W3CColourDatabase
#-------------------------------------------------------------------------

class W3CColourDatabase(object):
    """ Proxy for the ColourDatabase which allows for finding W3C colors.

        This class is necessary because the wx 'green' is the W3C 'lime',
        and we need some means to lookup the color names since wx has
        only a few hardcoded.

        This class is a proxy because AddColour expects a wx.ColourDatabase
        instance, not an instance of a subclass
    """

    _database = wx.ColourDatabase()

    def __init__(self):
        self._color_names = ['aqua', 'black', 'blue', 'fuchsia', 'gray',
                             'green', 'lime', 'maroon', 'navy', 'olive',
                             'purple', 'red', 'silver', 'teal', 'white',
                             'yellow']

        self.AddColour('aqua', wx.Colour(0, 0xff, 0xff, 255))
        self.AddColour('fuchsia', wx.Colour(0xff, 0, 0xff, 255))
        self.AddColour('green', wx.Colour(0, 0x80, 0, 255))
        self.AddColour('lime', wx.Colour(0, 0xff, 0, 255))
        self.AddColour('maroon', wx.Colour(0x80, 0x0, 0, 255))
        self.AddColour('navy', wx.Colour(0x00, 0x0, 0x80, 255))
        self.AddColour('olive', wx.Colour(0x80, 0x80, 0, 255))
        self.AddColour('purple', wx.Colour(0x80, 0x00, 0x80, 255))
        self.AddColour('silver', wx.Colour(0xc0, 0xc0, 0xc0, 255))
        self.AddColour('teal', wx.Colour(0, 0x80, 0x80, 255))

    def AddColour(self, name, color):
        if name not in self._color_names:
            self._color_names.append(name)
        return self._database.AddColour(name, color)

    def Find(self, color_name):
        return self._database.Find(color_name)

    def FindName(self, color):
        for color_name in self._color_names:
            if self.Find(color_name) == color:
                return color_name

        return ''

w3c_color_database = W3CColourDatabase()

#-------------------------------------------------------------------------
#  Convert a number into a wxColour object:
#-------------------------------------------------------------------------


def tuple_to_wxcolor(tup):
    if 3 <= len(tup) <= 4:
        for c in tup:
            if not isinstance(c, int):
                raise TraitError

        return wx.Colour(*tup)
    else:
        raise TraitError


def convert_to_color(object, name, value):
    """ Converts a number into a wxColour object.
    """
    if isinstance(value, tuple):
        return tuple_to_wxcolor(value)

    elif isinstance(value, ColourPtr):
        return wx.Colour(value.Red(), value.Green(), value.Blue())

    elif isinstance(value, wx.Colour):
        return value

    elif isinstance(value, str):

        if value in standard_colors:
            return standard_colors[value]

        # Check for tuple-ness
        tmp = value.strip()
        if tmp.startswith("(") and tmp.endswith(
                ")") and tmp.count(",") in (2, 3):
            tup = eval(tmp)
            return tuple_to_wxcolor(tup)

    elif isinstance(value, int):
        num = int(value)
        return wx.Colour(num / 0x10000, (num / 0x100) & 0xFF, num & 0xFF)

    raise TraitError

convert_to_color.info = ('a string of the form (r,g,b) or (r,g,b,a) where r, '
                         'g, b, and a are integers from 0 to 255, a wx.Colour '
                         'instance, an integer which in hex is of the form '
                         '0xRRGGBB, where RR is red, GG is green, and BB is '
                         'blue')

#-------------------------------------------------------------------------
#  Standard colors:
#-------------------------------------------------------------------------

standard_colors = {}
for name in ['aquamarine', 'black', 'blue', 'blue violet', 'brown',
             'cadet blue', 'coral', 'cornflower blue', 'cyan', 'dark grey',
             'dark green', 'dark olive green', 'dark orchid',
             'dark slate blue', 'dark slate grey', 'dark turquoise',
             'dim grey', 'firebrick', 'forest green', 'gold', 'goldenrod',
             'grey', 'green', 'green yellow', 'indian red', 'khaki',
             'light blue', 'light grey', 'light steel blue', 'lime green',
             'magenta', 'maroon', 'medium aquamarine', 'medium blue',
             'medium forest green', 'medium goldenrod', 'medium orchid',
             'medium sea green', 'medium slate blue', 'medium spring green',
             'medium turquoise', 'medium violet red', 'midnight blue', 'navy',
             'orange', 'orange red', 'orchid', 'pale green', 'pink', 'plum',
             'purple', 'red', 'salmon', 'sea green', 'sienna', 'sky blue',
             'slate blue', 'spring green', 'steel blue', 'tan', 'thistle',
             'turquoise', 'violet', 'violet red', 'wheat', 'white', 'yellow',
             'yellow green']:
    try:
        wx_color = w3c_color_database.Find(name)
        standard_colors[name] = convert_to_color(None, None,
                                                 wx_color)
    except:
        pass

#-------------------------------------------------------------------------
#  Define wxPython specific color traits:
#-------------------------------------------------------------------------

### Note: Declare the editor to be a function which returns the ColorEditor
# class from traits ui to avoid circular import issues. For backwards
# compatibility with previous Traits versions, the 'editors' folder in Traits
# project declares 'from api import *' in its __init__.py. The 'api' in turn
# can contain classes that have a Color trait which lead to this file getting
# imported. This leads to a circular import when declaring a Color trait.


def get_color_editor(*args, **traits):
    from color_editor import ToolkitEditorFactory
    return ToolkitEditorFactory(*args, **traits)


def WxColor(default='white', allow_none=False, **metadata):
    """ Defines wxPython-specific color traits.
    """

    if allow_none:
        return Trait(default, None, standard_colors, convert_to_color,
                     editor=get_color_editor, **metadata)

    return Trait(default, standard_colors, convert_to_color,
                 editor=get_color_editor, **metadata)



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enth373ought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
""" The 'null' specific implementations of the action manager internal classes.
"""


# Enthought library imports.
from traits.api import Any, Bool, HasTraits


class _MenuItem(HasTraits):
    """ A menu item representation of an action item. """

    #### '_MenuItem' interface ################################################

    # Is the item checked?
    checked = Bool(False)

    # A controller object we delegate taking actions through (if any).
    controller = Any

    # Is the item enabled?
    enabled = Bool(True)

    # Is the item visible?
    visible = Bool(True)

    # The radio group we are part of (None if the menu item is not part of such
    # a group).
    group = Any

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, parent, menu, item, controller):
        """ Creates a new menu item for an action item. """

        self.item = item

        self.control_id = 1
        self.control = None

        if controller is not None:
            self.controller = controller
            controller.add_to_menu(self)

        return


class _Tool(HasTraits):
    """ A tool bar tool representation of an action item. """

    #### '_Tool' interface ####################################################

    # Is the item checked?
    checked = Bool(False)

    # A controller object we delegate taking actions through (if any).
    controller = Any

    # Is the item enabled?
    enabled = Bool(True)

    # Is the item visible?
    visible = Bool(True)

    # The radio group we are part of (None if the tool is not part of such a
    # group).
    group = Any

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, parent, tool_bar, image_cache, item, controller,
                 show_labels):
        """ Creates a new tool bar tool for an action item. """

        self.item = item
        self.tool_bar = tool_bar

        # Create an appropriate tool depending on the style of the action.
        action  = self.item.action

        # If the action has an image then convert it to a bitmap (as required
        # by the toolbar).
        if action.image is not None:
            image = action.image.create_image()
            path = action.image.absolute_path
            bmp  = image_cache.get_bitmap(path)

        else:
            from pyface.api import ImageResource
            image = ImageResource('foo')
            bmp  = image.create_bitmap()

        self.control_id = 1
        self.control = None
        if controller is not None:
            self.controller = controller
            controller.add_to_toolbar(self)

        return


class _PaletteTool(HasTraits):
    """ A tool palette representation of an action item. """

    #### '_PaletteTool' interface #############################################

    # The radio group we are part of (None if the tool is not part of such a
    # group).
    group = Any

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, tool_palette, image_cache, item, show_labels):
        """ Creates a new tool palette tool for an action item. """

        self.item = item
        self.tool_palette = tool_palette


#### EOF ######################################################################



from __future__ import (absolute_import, division, print_function)

import numpy as np

from pyugrid.util import point_in_tri


def test_point_in_tri():
    test_datasets = [
        {
            'triangle': np.array([[0., 0.], [1., 0.], [0., 1.]]),
            'points_inside': [np.array([0.1, 0.1]), np.array([0.3, 0.3])],
            'points_outside': [np.array([5., 5.])],
        },
    ]

    for dataset in test_datasets:
        for point in dataset['points_inside']:
            assert point_in_tri(dataset['triangle'], point)
        for point in dataset['points_outside']:
            assert ~point_in_tri(dataset['triangle'], point)



"""
fs.contrib.bigfs
================

A FS object that represents the contents of a BIG file 
(C&C Generals, BfME C&C3, C&C Red Alert 3, C&C4 file format)

Written by Koen van de Sande
http://www.tibed.net

Contributed under the terms of the BSD License:
http://www.opensource.org/licenses/bsd-license.php
"""

from struct import pack, unpack

from fs.base import *
from fs.memoryfs import MemoryFS
from fs.filelike import StringIO

from fs.contrib.bigfs.subrangefile import SubrangeFile

class BIGEntry:
    def __init__(self, filename, offset, storedSize, isCompressed, realSize):
        self.filename = filename
        self.offset = offset
        self.storedSize = storedSize
        self.realSize = realSize
        self.isCompressed = isCompressed

    def getfile(self, baseFile):
        f = SubrangeFile(baseFile, self.offset, self.storedSize)
        if not self.isCompressed:
            return f
        else:
            return self.decompress(f, wrapAsFile=True)
    
    def getcontents(self, baseFile):
        f = SubrangeFile(baseFile, self.offset, self.storedSize)
        if not self.isCompressed:
            return f.read()
        else:
            return self.decompress(f, wrapAsFile=False)
    
    def decompress(self, g, wrapAsFile=True):
        buf = g.read(2)
        magic = unpack(">H", buf)[0]
        if (magic & 0x3EFF) == 0x10FB:
            # it is compressed
            if magic & 0x8000:
                outputSize = unpack(">I", g.read(4))[0]
                if magic & 0x100:
                    unknown1 = unpack(">I", g.read(4))[0]
            else:
                outputSize = unpack(">I", "\0" + g.read(3))[0]
                if magic & 0x100:
                    unknown1 = unpack(">I", "\0" + g.read(3))[0]
        
        output = []
        while True:
            opcode = unpack("B", g.read(1))[0]
            if not (opcode & 0x80):       # opcode: bit7==0 to get here
                # read second opcode
                opcode2 = unpack("B", g.read(1))[0]
                #print "0x80", toBits(opcode), toBits(opcode2), opcode & 0x03, (((opcode & 0x60) << 3) | opcode2) + Q, ((opcode & 0x1C) >> 2) + 2 + R
    
                # copy at most 3 bytes to output stream (lowest 2 bits of opcode)
                count = opcode & 0x03
                for i in range(count):
                    output.append(g.read(1))
                
                # you always have to look at least one byte, hence the +1
                # use bit6 and bit5 (bit7=0 to trigger the if-statement) of opcode, and 8 bits of opcode2 (10-bits)
                lookback = (((opcode & 0x60) << 3) | opcode2) + 1
                
                # use bit4..2 of opcode
                count = ((opcode & 0x1C) >> 2) + 3
                
                for i in range(count):
                    output.append(output[-lookback])
            elif not (opcode & 0x40):     # opcode: bit7..6==10 to get here
                opcode2 = unpack("B", g.read(1))[0]
                opcode3 = unpack("B", g.read(1))[0]
                #print "0x40", toBits(opcode), toBits(opcode2), toBits(opcode3)
                
                # copy count bytes (upper 2 bits of opcode2)
                count = opcode2 >> 6
                for i in range(count):
                    output.append(g.read(1))
                
                # look back again (lower 6 bits of opcode2, all 8 bits of opcode3, total 14-bits)
                lookback = (((opcode2 & 0x3F) << 8) | opcode3) + 1
                # lower 6 bits of opcode are the count to copy
                count = (opcode & 0x3F) + 4
                
                for i in range(count):
                    output.append(output[-lookback])
            elif not (opcode & 0x20):     # opcode: bit7..5=110 to get here
                opcode2 = unpack("B", g.read(1))[0]
                opcode3 = unpack("B", g.read(1))[0]
                opcode4 = unpack("B", g.read(1))[0]

                # copy at most 3 bytes to output stream (lowest 2 bits of opcode)
                count = opcode & 0x03
                for i in range(count):
                    output.append(g.read(1))
                
                # look back: bit4 of opcode, all bits of opcode2 and opcode3, total 17-bits
                lookback = (((opcode & 0x10) >> 4) << 16) | (opcode2 << 8) | (opcode3) + 1
                # bit3..2 of opcode and the whole of opcode4
                count = (((((opcode & 0x0C) >> 2) << 8)) | opcode4) + 5

                #print "0x20", toBits(opcode), toBits(opcode2), toBits(opcode3), toBits(opcode4), lookback, count

                for i in range(count):
                    output.append(output[-lookback])
            else:                         # opcode: bit7..5==1 to get here
                # use lowest 5 bits for count
                count = ((opcode & 0x1F) << 2) + 4
                if count > 0x70:   # this is end of input
                    # turn into a small-copy
                    count = opcode & 0x03
                    #print "0xEXITCOPY", count
                    for i in range(count):
                        output.append(g.read(1))
                    break

                # "big copy" operation: up to 112 bytes (minumum of 4, multiple of 4)
                for i in range(count):
                    output.append(g.read(1))
                #print "0xLO", toBits(opcode), count
        
        if wrapAsFile:
            return StringIO("".join(output))
        else:
            return "".join(output)
        
    def __str__(self):
        return "<BIGEntry %s offset=%d storedSize=%d isCompressed=%s realSize=%d in %s" % (self.filename, self.offset, self.storedSize, str(self.isCompressed), self.realSize, self.filenameBIG)
        

class _ExceptionProxy(object):

    """A placeholder for an object that may no longer be used."""

    def __getattr__(self, name):
        raise ValueError("File has been closed")

    def __setattr__(self, name, value):
        raise ValueError("File has been closed")

    def __nonzero__(self):
        return False


class BigFS(FS):

    """A FileSystem that represents a BIG file."""
    
    _meta = { 'virtual' : False,
              'read_only' : True,
              'unicode_paths' : True,
              'case_insensitive_paths' : False,
              'network' : False,                        
             }

    def __init__(self, filename, mode="r", thread_synchronize=True):
        """Create a FS that maps on to a big file.

        :param filename: A (system) path, or a file-like object
        :param mode: Mode to open file: 'r' for reading, 'w' and 'a' not supported
        :param thread_synchronize: -- Set to True (default) to enable thread-safety

        """
        super(BigFS, self).__init__(thread_synchronize=thread_synchronize)

        if len(mode) > 1 or mode not in "r":
            raise ValueError("mode must be 'r'")
        self.file_mode = mode
        self.big_path = str(filename)

        self.entries = {}
        try:
            self.bf = open(filename, "rb")
        except IOError:
            raise ResourceNotFoundError(str(filename), msg="BIG file does not exist: %(path)s")

        self._path_fs = MemoryFS()
        if mode in 'ra':
            self._parse_resource_list(self.bf)

    def __str__(self):
        return "<BigFS: %s>" % self.big_path

    def __unicode__(self):
        return unicode(self.__str__())


    def _parse_resource_list(self, g):
        magicWord = g.read(4)
        if magicWord != "BIGF" and magicWord != "BIG4":
            raise ValueError("Magic word of BIG file invalid: " + filename + " " + repr(magicWord))
        header = g.read(12)
        header = unpack(">III", header)
        BIGSize = header[0]
        fileCount = header[1]
        bodyOffset = header[2]
        for i in range(fileCount):
            fileHeader = g.read(8)
            fileHeader = unpack(">II", fileHeader)

            pos = g.tell()
            buf = g.read(4096)
            marker = buf.find("\0")
            if marker == -1:
                raise ValueError("Could not parse filename in BIG file: Too long or invalid file")
            name = buf[:marker]
            # TODO: decode the encoding of name (or normalize the path?)
            isCompressed, uncompressedSize = self.__isCompressed(g, fileHeader[0], fileHeader[1])
            be = BIGEntry(name, fileHeader[0], fileHeader[1], isCompressed, uncompressedSize)
            name = normpath(name)
            self.entries[name] = be
            self._add_resource(name)
            g.seek(pos + marker + 1)

    def __isCompressed(self, g, offset, size):
        g.seek(offset)
        buf = g.read(2)
        magic = unpack(">H", buf)[0]
        if (magic & 0x3EFF) == 0x10FB:
            # it is compressed
            if magic & 0x8000:
                # decompressed size is uint32
                return True, unpack(">I", g.read(4))[0]
            else:
                # use only 3 bytes
                return True, unpack(">I", "\0" + g.read(3))[0]
        return False, size

    def _add_resource(self, path):
        if path.endswith('/'):
            path = path[:-1]
            if path:
                self._path_fs.makedir(path, recursive=True, allow_recreate=True)
        else:
            dirpath, filename = pathsplit(path)
            if dirpath:
                self._path_fs.makedir(dirpath, recursive=True, allow_recreate=True)
            f = self._path_fs.open(path, 'w')
            f.close()


    def close(self):
        """Finalizes the zip file so that it can be read.
        No further operations will work after this method is called."""

        if hasattr(self, 'bf') and self.bf:
            self.bf.close()
            self.bf = _ExceptionProxy()

    @synchronize
    def open(self, path, mode="r", **kwargs):
        path = normpath(relpath(path))        

        if 'r' in mode:
            if self.file_mode not in 'ra':
                raise OperationFailedError("open file", path=path, msg="Big file must be opened for reading ('r') or appending ('a')")
            try:
                return self.entries[path].getfile(self.bf)
            except KeyError:
                raise ResourceNotFoundError(path)

        if 'w' in mode:
            raise OperationFailedError("open file", path=path, msg="Big file cannot be edited ATM")

        raise ValueError("Mode must contain be 'r' or 'w'")

    @synchronize
    def getcontents(self, path):
        if not self.exists(path):
            raise ResourceNotFoundError(path)
        path = normpath(path)
        try:
            contents = self.entries[path].getcontents(self.bf)
        except KeyError:
            raise ResourceNotFoundError(path)
        except RuntimeError:
            raise OperationFailedError("read file", path=path, msg="Big file must be oppened with 'r' or 'a' to read")
        return contents

    def desc(self, path):
        if self.isdir(path):
            return "Dir in big file: %s" % self.big_path
        else:
            return "File in big file: %s" % self.big_path

    def isdir(self, path):
        return self._path_fs.isdir(path)

    def isfile(self, path):
        return self._path_fs.isfile(path)

    def exists(self, path):
        return self._path_fs.exists(path)

    @synchronize
    def makedir(self, dirname, recursive=False, allow_recreate=False):
        dirname = normpath(dirname)
        if self.file_mode not in "wa":
            raise OperationFailedError("create directory", path=dirname, msg="Big file must be opened for writing ('w') or appending ('a')")
        if not dirname.endswith('/'):
            dirname += '/'
        self._add_resource(dirname)

    def listdir(self, path="/", wildcard=None, full=False, absolute=False, dirs_only=False, files_only=False):
        return self._path_fs.listdir(path, wildcard, full, absolute, dirs_only, files_only)

    @synchronize
    def getinfo(self, path):
        if not self.exists(path):
            raise ResourceNotFoundError(path)
        path = normpath(path).lstrip('/')
        info = {'size': 0}
        if path in self.entries:
            be = self.entries[path]
            info['size'] = be.realSize
            info['file_size'] = be.realSize
            info['stored_size'] = be.storedSize
            info['is_compressed'] = be.isCompressed
            info['offset'] = be.offset
            info['internal_filename'] = be.filename
            info['filename'] = path
        return info



"""
Decorator to mark functions and methods as recordable.
"""
# Author: Prabhu Ramachandran <prabhu@aero.iitb.ac.in>
# Copyright (c) 2008, Prabhu Ramachandran and Enthought, Inc.
# License: BSD Style.

from .package_globals import get_recorder


# Guard to ensure that only the outermost recordable call is recorded
# and nested calls ignored.
_outermost_call = True


def recordable(func):
    """A decorator that wraps a function into one that is recordable.

    This will record the function only if the global recorder has been
    set via a `set_recorder` function call.

    This is almost entirely copied from the
    apptools.appscripting.scriptable.scriptable decorator.
    """

    def _wrapper(*args, **kw):
        """A wrapper returned to replace the decorated function."""
        global _outermost_call

        # Boolean to specify if the method was recorded or not.
        record = False
        if _outermost_call:
            # Get the recorder.
            rec = get_recorder()
            if rec is not None:
                _outermost_call = False
                # Record the method if recorder is available.
                record = True
                try:
                    result = rec.record_function(func, args, kw)
                finally:
                    _outermost_call = True
        if not record:
            # If the method was not recorded, just call it.
            result = func(*args, **kw)

        return result

    # Mimic the actual function.
    _wrapper.__name__ = func.__name__
    _wrapper.__doc__ = func.__doc__
    _wrapper.__dict__.update(func.__dict__)

    return _wrapper



from traits.api import HasTraits, List, provides, Str

from apptools.selection.i_selection import IListSelection


@provides(IListSelection)
class ListSelection(HasTraits):
    """ Selection for ordered sequences of items.

    This is the default implementation of the :class:`~.IListSelection`
    interface.
    """

    #### 'ISelection' protocol ################################################

    #: ID of the selection provider that created this selection object.
    provider_id = Str

    def is_empty(self):
        """ Is the selection empty? """
        return len(self.items) == 0

    #### 'IListSelection' protocol ############################################

    #: Selected objects.
    items = List

    #: Indices of the selected objects in the selection provider.
    indices = List

    #### 'ListSelection' class protocol #######################################

    @classmethod
    def from_available_items(cls, provider_id, selected, all_items):
        """ Create a list selection given a list of all available items.

        Fills in the required information (in particular, the indices) based
        on a list of selected items and a list of all available items.

        .. note::
            - The list of available items must not contain any duplicate items.
            - It is expected that ``selected`` is populated by items in
              ``all_items``.

        """
        number_of_items = len(all_items)
        indices = []

        for item in selected:
            for index in range(number_of_items):
                if all_items[index] is item:
                    indices.append(index)
                    break
            else:
                msg = 'Selected item: {!r}, could not be found'
                raise ValueError(msg.format(item))

        return cls(provider_id=provider_id, items=selected, indices=indices)



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
""" The interface for a dialog that displays a message. """


# Enthought library imports.
from traits.api import Enum, Unicode

# Local imports.
from pyface.i_dialog import IDialog


class IMessageDialog(IDialog):
    """ The interface for a dialog that displays a message. """

    #### 'IMessageDialog' interface ###########################################

    #: The message to display in the dialog.
    message = Unicode

    #: More information about the message to be displayed.
    informative = Unicode

    #: More detail about the message to be displayed in the dialog.
    detail = Unicode

    #: The severity of the message.
    severity = Enum('information', 'warning', 'error')


class MMessageDialog(object):
    """ The mixin class that contains common code for toolkit specific
    implementations of the IMessageDialog interface.
    """



from traits.api import HasTraits, Instance, Str
from traitsui.item import Item
from traitsui.view import View
from traitsui.tests._tools import (
    press_ok_button, skip_if_not_qt4, store_exceptions_on_all_threads)


class EditedInstance(HasTraits):
    value = Str
    traits_view = View(Item('value'), buttons=['OK'])


class NonmodalInstanceEditor(HasTraits):
    inst = Instance(EditedInstance, ())
    traits_view = View(Item('inst', style='simple'), buttons=['OK'])


@skip_if_not_qt4
def test_simple_editor():
    with store_exceptions_on_all_threads():
        obj = NonmodalInstanceEditor()
        ui = obj.edit_traits()
        editor = ui.get_editors('inst')[0]

        # make the dialog appear
        editor._button.click()

        # close the ui dialog
        press_ok_button(editor._dialog_ui)

        # close the main ui
        press_ok_button(ui)


@skip_if_not_qt4
def test_simple_editor_parent_closed():
    with store_exceptions_on_all_threads():
        obj = NonmodalInstanceEditor()
        ui = obj.edit_traits()
        editor = ui.get_editors('inst')[0]

        # make the dialog appear
        editor._button.click()

        # close the main ui
        press_ok_button(ui)
